package cn.iocoder.yudao.module.product.service.product.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import cn.iocoder.yudao.framework.apollo.core.dto.ChannelClearancePriceDto;
import cn.iocoder.yudao.framework.apollo.core.event.ChannelPriceEvent;
import cn.iocoder.yudao.framework.apollo.core.event.ProductLineUpdateAutoProcessOverWeightExceptionEvent;
import cn.iocoder.yudao.framework.common.enums.CommonStatusEnum;
import cn.iocoder.yudao.framework.common.exception.ServiceException;
import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.dict.core.dto.DictDataRespDTO;
import cn.iocoder.yudao.framework.mybatis.core.query.LambdaQueryWrapperX;
import cn.iocoder.yudao.framework.mybatis.core.service.AbstractService;
import cn.iocoder.yudao.framework.mybatis.core.util.MyBatisUtils;
import cn.iocoder.yudao.framework.security.core.util.SecurityFrameworkUtils;
import cn.iocoder.yudao.module.depository.dal.dataobject.warehouse.WarehouseLineDO;
import cn.iocoder.yudao.module.depository.dal.mysql.warehouse.WarehouseLineMapper;
import cn.iocoder.yudao.module.depository.service.warehouse.WarehouseService;
import cn.iocoder.yudao.module.depository.vo.warehouse.WarehouseLineSearchVO;
import cn.iocoder.yudao.module.ecw.api.currency.CurrencyApi;
import cn.iocoder.yudao.module.ecw.api.currency.dto.ExchangeRateRespDTO;
import cn.iocoder.yudao.module.ecw.api.internalMessage.InternalMessageApi;
import cn.iocoder.yudao.module.ecw.api.internalMessage.dto.InternalMessageCreateDto;
import cn.iocoder.yudao.module.product.convert.product.ProductPriceConvert;
import cn.iocoder.yudao.module.product.dal.dataobject.product.*;
import cn.iocoder.yudao.module.product.dal.dataobject.productPriceSnapshot.ProductPriceSnapshotDO;
import cn.iocoder.yudao.module.product.dal.mysql.product.*;
import cn.iocoder.yudao.module.product.dal.mysql.productPriceSnapshot.ProductPriceSnapshotMapper;
import cn.iocoder.yudao.module.product.dto.AirClearancePriceResultDto;
import cn.iocoder.yudao.module.product.dto.CopyProductPriceDto;
import cn.iocoder.yudao.module.product.dto.ProductPriceExcelData;
import cn.iocoder.yudao.module.product.enums.PriceTypeEnum;
import cn.iocoder.yudao.module.product.param.PriceParam;
import cn.iocoder.yudao.module.product.service.product.*;
import cn.iocoder.yudao.framework.apollo.core.event.ChannelDataEvent;
import cn.iocoder.yudao.module.product.vo.productPrice.*;
import cn.iocoder.yudao.module.system.api.dict.DictDataApi;
import cn.iocoder.yudao.module.system.api.user.AdminUserApi;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.StringUtils;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.google.common.base.Joiner;
import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.quartz.SchedulerException;
import org.redisson.api.*;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.Duration;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.PRODUCT_PRICE_NOT_EXISTS;
import static java.util.stream.Collectors.toList;

/**
 * 产品价格 Service 实现类
 *
 * @author wlh
 */
@Slf4j
@Service
@Validated
public class ProductPriceServiceImpl extends AbstractService<ProductPriceMapper, ProductPriceDO> implements ProductPriceService {

    @Resource
    private ProductPriceMapper productPriceMapper;
    @Resource
    private ProductPriceSpecialMapper productPriceSpecialMapper;
    @Resource
    private ProductPricePackagingMapper productPricePackagingMapper;
    @Resource
    private ProductPriceStepMapper productPriceStepMapper;

    @Resource
    private ProductPriceStepSpecialMapper productPriceStepSpecialMapper;
    @Resource
    private ProductPriceStepPackagingMapper productPriceStepPackagingMapper;

    @Autowired
    private ApplicationContext applicationContext;
    @Resource
    private RedissonClient redissonClient;
    @Resource
    private WarehouseLineMapper warehouseLineMapper;

    @Resource
    private WarehouseService warehouseService;

    @Resource
    private ProductBatchLogService productBatchLogService;

    @Resource
    private InternalMessageApi internalMessageApi;

    @Resource
    private ProductService productService;
    @Resource
    private ProductPriceSpecialService productPriceSpecialService;
    @Resource
    private ProductPriceStepService productPriceStepService;
    @Resource
    private AdminUserApi adminUserApi;
    @Resource
    private CurrencyApi currencyApi;
    @Resource
    private ProductPriceSnapshotMapper productPriceSnapshotMapper;
    @Resource
    private DictDataApi dictDataApi;
    @Resource
    private ProductPriceStepClearanceService productPriceStepClearanceService;


    private static final String PREFIX = "price_batch_update_";
    private static final Integer DATA_THRESHOLD = 50;

    private static final String BATCH_DEL_SCRIPT = "for k,v in  pairs(ARGV) do \n " +
            "redis.call('del',string.gsub(v,'\"',''),'') \n" +
            "end \n" +
            "return #ARGV";
    //redissonClient.getScript().evalAsync(RScript.Mode.READ_WRITE, BATCH_DEL_SCRIPT, RScript.ReturnType.INTEGER, keys, keys.toArray());

    @PostConstruct
    public void init() throws SchedulerException {
        log.info("清空路线价格缓存数据");
        redissonClient.getKeys().deleteByPattern(PREFIX + "*");
    }

    @Override
    public void clearKey() {
        redissonClient.getKeys().deleteByPattern(PREFIX + "*");
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Long singleCreate(ProductPriceCreateReqVO createReqVO) {
        List<LineChannelVo> lineChannelList = createReqVO.getLineChannelList();
        ProductDO productDO = checkUpdateStatus(lineChannelList, createReqVO.getProductId());

        List<ProductDO> productList = Collections.singletonList(productDO);
        //商品线路保存在redis。更新成功后移除
        batchSaveKeyInCache(productList, lineChannelList);

        WarehouseLineSearchVO searchVO = new WarehouseLineSearchVO();
        List<WarehouseLineDO> warehouseLineList = warehouseService.openedRouterList(searchVO);
        createReqVO.setSync(true);
        createReqVO.setBatch(false);
        batchUpdate(createReqVO, warehouseLineList, productList);
        return 1L;
    }

    @Override
    public Long singleCreateAir(ProductPriceAirCreateReqVO createReqVO) {
        List<LineChannelVo> lineChannelList = createReqVO.getLineChannelList();
        ProductDO productDO = checkUpdateStatus(lineChannelList, createReqVO.getProductId());

        WarehouseLineSearchVO searchVO = new WarehouseLineSearchVO();
        List<WarehouseLineDO> warehouseLineList = warehouseService.openedRouterList(searchVO);


        List<ProductDO> productList = Collections.singletonList(productDO);
        //商品线路保存在redis。更新成功后移除
        batchSaveKeyInCache(productList, lineChannelList);

        createReqVO.setSync(true);
        createReqVO.setBatch(false);
        batchUpdateAir(createReqVO, warehouseLineList, productList);
        return 1L;
    }

    private ProductDO checkUpdateStatus(List<LineChannelVo> lineChannelList, Long productId) {
        if (CollectionUtil.isEmpty(lineChannelList)) {
            throw new IllegalArgumentException("线路渠道列表不能为空");
        }
        ProductDO productDO = productService.getProduct(productId);
        if (productDO.getAuditStatus() != 2) {
            throw new IllegalArgumentException("商品未审核通过");
        }
        //检查是否在更新中
        check(lineChannelList, productId);
        return productDO;
    }

    private void check(List<LineChannelVo> lineChannelList, Long productId) {
        ProductPriceReqBaseVO productPriceReqBaseVO = new ProductPriceAirBaseVO();
        productPriceReqBaseVO.setLineChannelList(lineChannelList);
        productPriceReqBaseVO.setIsAllProduct(0);
        productPriceReqBaseVO.setProductIdList(Collections.singletonList(productId));
        check(productPriceReqBaseVO);
    }

    @Transactional(rollbackFor = Exception.class)
    public void batchUpdate(ProductPriceCreateReqVO createReqVO, List<WarehouseLineDO> warehouseLineList, List<ProductDO> productList) {
        long t1 = System.currentTimeMillis();
        Long loginUserId = SecurityFrameworkUtils.getLoginUserId();


        ProductPriceDO productPrice = ProductPriceConvert.INSTANCE.convert(createReqVO);

        if (null != createReqVO.getStepPrice() && createReqVO.getStepPrice() == 1) {
            checkAndUpdateStepToMainPrice(productPrice, createReqVO);
        }

        List<LineChannelVo> lineChannelList = createReqVO.getLineChannelList();
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());

        Map<Long, String> lineTransportTypeMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, WarehouseLineDO::getTransportType));

        List<Long> idList = productList.stream()
                .map(ProductDO::getId)
                .collect(toList());
        //这个产品这么多线路已有的价格
        List<ProductPriceDO> dbPriceList = productPriceMapper.selectList(new LambdaQueryWrapperX<ProductPriceDO>()
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
                .in(ProductPriceDO::getProductId, idList));
        Map<String, ProductPriceDO> productLineChannelMap = dbPriceList.stream()
                .collect(Collectors.toMap(t -> t.getProductId() + "-" + t.getWarehouseLineId() + "-" + t.getShippingChannelId(), v -> v, (v1, v2) -> v2));


        List<ProductPriceDO> insertList = new ArrayList<>();
        List<ProductPriceDO> updateList = new ArrayList<>();

        List<ProductPriceDO> snapshotList = new ArrayList<>();

        Date current = new Date();
        for (Long productId : idList) {
            for (LineChannelVo lineChannel : lineChannelList) {
                Long lineId = lineChannel.getLineId();
                Long shippingChannelId = lineChannel.getShippingChannelId();
                String key = productId + "-" + lineId + "-" + shippingChannelId;

                ProductPriceDO dbItem;
                if (productLineChannelMap.containsKey(key)) {
                    dbItem = productLineChannelMap.get(key);

                    //添加日志
                    ProductPriceDO snapshot = ObjectUtil.clone(dbItem);
                    snapshotList.add(snapshot);
                    dbItem.setUpdater(String.valueOf(loginUserId));
                    updateColumn(dbItem, productPrice);
                    updateList.add(dbItem);
                } else {
                    dbItem = ObjectUtil.clone(productPrice);
                    dbItem.setProductId(productId);
                    dbItem.setWarehouseLineId(lineId);
                    dbItem.setShippingChannelId(shippingChannelId);

                    if (dbItem.getNeedBook() == null) {
                        dbItem.setNeedBook(0);
                    }
                    if (dbItem.getPriceType() == null) {
                        dbItem.setPriceType(0);
                    }
                    if (dbItem.getNeedPay() == null) {
                        dbItem.setNeedPay(0);
                    }
                    if (dbItem.getAuditStatus() == null) {
                        dbItem.setAuditStatus(0);
                    }
                    if (dbItem.getPayDelivery() == null) {
                        dbItem.setPayDelivery(0);
                    }
                    if (dbItem.getStepPrice() == null) {
                        dbItem.setStepPrice(0);
                    }
                    if (dbItem.getMinWeight() == null) {
                        dbItem.setMinWeight(BigDecimal.ZERO);
                    }
                    if (dbItem.getMinWeightUnit() == null) {
                        dbItem.setMinWeightUnit(7L);
                    }
                    dbItem.setBlacklist(0);
                    dbItem.setStatus(0);
                    dbItem.setCreateTime(current);
                    insertList.add(dbItem);
                }
                dbItem.setUpdateTime(current);
                String transportType = lineTransportTypeMap.get(lineId);
                initBasePrice(dbItem, transportType);
            }
        }
        if (CollectionUtil.isNotEmpty(insertList)) {
            productPriceMapper.insertBatchSomeColumn(insertList);
        }
        if (CollectionUtil.isNotEmpty(updateList)) {
            productPriceMapper.updateBatch(updateList);
        }
        List<ProductPriceDO> productPriceList = new ArrayList<>();
        productPriceList.addAll(insertList);
        productPriceList.addAll(updateList);
        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        productPriceList.forEach(t -> priceTransportTypeMap.put(t.getId(), lineTransportTypeMap.get(t.getWarehouseLineId())));

        //添加修改日志,需要查询特需和阶梯，所以需要在修改之前执行
        batchAddSnapshot(snapshotList);

        //处理特需价格
        List<Long> productPriceIdList = productPriceList.stream().map(ProductPriceDO::getId).collect(Collectors.toList());
        clearAndSaveSpecialList(productPriceIdList, createReqVO.getSpecialList(), priceTransportTypeMap);

        //处理包装单位价格
        List<ProductPricePackagingDO> allPackageList = buildAllPackageList(createReqVO);
        clearAndSavePackagingList(productPriceIdList, allPackageList, priceTransportTypeMap, createReqVO.isBatch());

        //判断是否是阶梯定价
        if (null != createReqVO.getStepPrice() && createReqVO.getStepPrice() == 1) {
            List<ProductPriceStepDO> allStepList = buildAllStepList(createReqVO);
            clearAndSavePriceStepList(productPriceIdList, allStepList, priceTransportTypeMap, createReqVO.isBatch());
        }
        long t2 = System.currentTimeMillis();
        log.info("路线处理耗时:" + (t2 - t1));
        deleteKey(idList, lineChannelList);

        handleOrderException(productPriceIdList, createReqVO.isSync());
    }

    /**
     * 批量添加快照日志
     *
     * @param snapshotList 快照列表
     */
    private void batchAddSnapshot(List<ProductPriceDO> snapshotList) {
        if (CollectionUtil.isEmpty(snapshotList)) return;

        Long loginUserId = SecurityFrameworkUtils.getLoginUserId();
        String creatorName = adminUserApi.getUserName(String.valueOf(loginUserId));

        int batchCount = 500;
        //sql长度有限制，分批次
        int countBatch = (int) Math.ceil((double) snapshotList.size() / batchCount);
        List<List<ProductPriceDO>> splitList = Stream.iterate(0, n -> n + 1)
                .limit(countBatch)
                .parallel()
                .map(a -> snapshotList.stream().skip(a * batchCount).limit(batchCount).parallel().collect(toList())
                ).collect(toList());
        for (List<ProductPriceDO> productPriceList : splitList) {
            List<Long> idList = productPriceList.stream()
                    .map(ProductPriceDO::getId)
                    .collect(toList());
            List<ProductPriceSpecialDO> specialList = productPriceSpecialService.list(new LambdaQueryWrapperX<ProductPriceSpecialDO>()
                    .in(ProductPriceSpecialDO::getProductPriceId, idList)
            );
            Map<Long, List<ProductPriceSpecialDO>> specialMap = specialList.stream()
                    .collect(Collectors.groupingBy(ProductPriceSpecialDO::getProductPriceId));

            List<ProductPriceStepDO> stepList = productPriceStepMapper.selectList(new LambdaQueryWrapper<ProductPriceStepDO>()
                    .in(ProductPriceStepDO::getProductPriceId, idList)
            );

            Map<Long, List<ProductPriceStepDO>> stepMap = stepList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepDO::getProductPriceId));

            List<ProductPriceSnapshotDO> priceSnapshotList = new ArrayList<>();
            for (ProductPriceDO productPriceDO : productPriceList) {
                Long priceId = productPriceDO.getId();
                Long productId = productPriceDO.getProductId();
                ProductPriceSnapshotDO snapshotDO = new ProductPriceSnapshotDO();
                snapshotDO.setPriceId(priceId);
                snapshotDO.setProductId(productId);
                snapshotDO.setCreatorName(creatorName);
                snapshotDO.setWarehouseLineId(productPriceDO.getWarehouseLineId());
                snapshotDO.setShippingChannelId(productPriceDO.getShippingChannelId());
                ProductPriceRespVO respVO = ProductPriceConvert.INSTANCE.convert(productPriceDO);
                respVO.setSpecialList(specialMap.get(priceId));
                List<ProductPriceStepDO> allStepLIst = stepMap.get(priceId);
                buildStepTypeList(allStepLIst, respVO);
                String content = JSON.toJSONString(respVO);
                snapshotDO.setContent(content);
                snapshotDO.setDeleted(false);
                priceSnapshotList.add(snapshotDO);
            }
            productPriceSnapshotMapper.insertBatchSomeColumn(priceSnapshotList);
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Long createProductPrice(ProductPriceCreateReqVO createReqVO, List<WarehouseLineDO> warehouseLineList) {
        // 插入
        ProductPriceDO productPrice = ProductPriceConvert.INSTANCE.convert(createReqVO);
        List<ProductPriceDO> productPriceList = new ArrayList<>();
        List<LineChannelVo> lineChannelList = createReqVO.getLineChannelList();
        Long productId = productPrice.getProductId();
        long t1 = System.currentTimeMillis();
        Map<Long, String> lineTransportTypeMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, WarehouseLineDO::getTransportType));

        for (LineChannelVo lineChannel : lineChannelList) {
            //已有这个价格则修改
            Long lineId = lineChannel.getLineId();

            Long shippingChannelId = lineChannel.getShippingChannelId();
            ProductPriceDO dbItem = productPriceMapper.selectOne(new LambdaQueryWrapperX<ProductPriceDO>()
                    .eq(ProductPriceDO::getWarehouseLineId, lineId)
                    .eqIfPresent(ProductPriceDO::getShippingChannelId, shippingChannelId)
                    .eq(ProductPriceDO::getProductId, productId));
            if (dbItem == null) {
                dbItem = ObjectUtil.clone(productPrice);
                dbItem.setWarehouseLineId(lineId);
                dbItem.setShippingChannelId(shippingChannelId);
            } else {
                //更新属性
                updateColumn(dbItem, productPrice);
            }

            String transportType = lineTransportTypeMap.get(lineId);

            if (dbItem.getTransportVolumeUnit() == null) {
                dbItem.setTransportVolumeUnit(getUnitByTransportType(transportType));
            }
            if (dbItem.getClearanceVolumeUnit() == null) {
                dbItem.setClearanceVolumeUnit(getUnitByTransportType(transportType));
            }
            if (dbItem.getAllVolumeUnit() == null) {
                dbItem.setAllVolumeUnit(getUnitByTransportType(transportType));
            }

            productPriceList.add(dbItem);
        }
        saveOrUpdateBatch(productPriceList);

        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        productPriceList.forEach(t -> priceTransportTypeMap.put(t.getId(), lineTransportTypeMap.get(t.getWarehouseLineId())));

        List<Long> productPriceIdList = productPriceList.stream().map(ProductPriceDO::getId).collect(Collectors.toList());
        clearAndSaveSpecialList(productPriceIdList, createReqVO.getSpecialList(), priceTransportTypeMap);

        //处理包装单位价格
        List<ProductPricePackagingDO> allPackageList = buildAllPackageList(createReqVO);
        clearAndSavePackagingList(productPriceIdList, allPackageList, priceTransportTypeMap, true);

        //判断是否是阶梯定价
        if (null != createReqVO.getStepPrice() && createReqVO.getStepPrice() == 1) {
            List<ProductPriceStepDO> allStepList = buildAllStepList(createReqVO);
            clearAndSavePriceStepList(productPriceIdList, allStepList, priceTransportTypeMap, true);
        }
        long t2 = System.currentTimeMillis();
        log.info("路线处理耗时:" + (t2 - t1));
        handleOrderException(productPriceIdList, true);
        deleteKey(Collections.singletonList(productId), lineChannelList);
        return productPrice.getId();
    }

    private void updateColumn(ProductPriceDO dbItem, ProductPriceDO productPrice) {
        if (Objects.nonNull(productPrice.getTransportPrice())) {
            dbItem.setTransportPrice(productPrice.getTransportPrice());
            dbItem.setTransportPriceUnit(productPrice.getTransportPriceUnit());
            dbItem.setTransportVolumeUnit(productPrice.getTransportVolumeUnit());
        }
        if (Objects.nonNull(productPrice.getClearancePrice())) {
            dbItem.setClearancePrice(productPrice.getClearancePrice());
            dbItem.setClearancePriceUnit(productPrice.getClearancePriceUnit());
            dbItem.setClearanceVolumeUnit(productPrice.getClearanceVolumeUnit());
        }

        if (Objects.nonNull(productPrice.getAllPrice())) {
            dbItem.setAllPrice(productPrice.getAllPrice());
            dbItem.setAllPriceUnit(productPrice.getAllPriceUnit());
            dbItem.setAllVolumeUnit(productPrice.getAllVolumeUnit());
        }
        if (Objects.nonNull(productPrice.getMinWeight())) {
            dbItem.setMinWeight(productPrice.getMinWeight());
        }
        if (Objects.nonNull(productPrice.getMinWeightUnit())) {
            dbItem.setMinWeightUnit(productPrice.getMinWeightUnit());
        }
        if(Objects.nonNull(productPrice.getWeightLimit())) {
            dbItem.setWeightLimit(productPrice.getWeightLimit());
        }

        if(Objects.nonNull(productPrice.getPriceType())) {
            //单价模式不为空则修改
            dbItem.setPriceType(productPrice.getPriceType());

            if (productPrice.getPriceType() == 0) {
                // 清关费+运费 清空全包价设置
                dbItem.setAllPrice(null);
                dbItem.setAllPriceUnit(null);
                dbItem.setAllVolumeUnit(null);
            }
            if (productPrice.getPriceType() == 1) {
                // 全包价 清空清关费+运费设置
                dbItem.setTransportPrice(null);
                dbItem.setTransportPriceUnit(null);
                dbItem.setTransportVolumeUnit(null);
                dbItem.setClearancePrice(null);
                dbItem.setClearancePriceUnit(null);
                dbItem.setClearanceVolumeUnit(null);
            }
        }

//        if (Objects.isNull(productPrice.getPriceType())) {
//            productPrice.setPriceType(dbItem.getPriceType());
//        }
//        if (productPrice.getPriceType() == 0) {
//            // 清关费+运费 清空全包价设置
//            dbItem.setAllPrice(null);
//            dbItem.setAllPriceUnit(null);
//            dbItem.setAllVolumeUnit(null);
//        }
//        if (productPrice.getPriceType() == 1) {
//            // 全包价 清空清关费+运费设置
//            dbItem.setTransportPrice(null);
//            dbItem.setTransportPriceUnit(null);
//            dbItem.setTransportVolumeUnit(null);
//            dbItem.setClearancePrice(null);
//            dbItem.setClearancePriceUnit(null);
//            dbItem.setClearanceVolumeUnit(null);
//        }
        dbItem.setValidateStartDate(productPrice.getValidateStartDate());
        dbItem.setValidateEndDate(productPrice.getValidateEndDate());

        dbItem.setNeedBook(productPrice.getNeedBook());
        dbItem.setDayLimit(productPrice.getDayLimit());
        dbItem.setContainerLocation(productPrice.getContainerLocation());
        dbItem.setSquare(productPrice.getSquare());
        dbItem.setNeedPay(productPrice.getNeedPay());
//        dbItem.setPriceType(productPrice.getPriceType());
        dbItem.setStepPrice(productPrice.getStepPrice());
    }

    private void deleteKey(List<Long> productIdList, List<LineChannelVo> lineChannelList) {
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());

        List<String> keyList = new ArrayList<>();
        for (Long productId : productIdList) {
            for (Long lineId : lineIdList) {
                String key = PREFIX + productId + "_" + lineId;
                keyList.add(key);
//                redissonClient.getBucket(key).deleteAsync();
            }
        }
        redissonClient.getKeys().delete(keyList.toArray(new String[0]));
    }

    private Long getUnitByTransportType(String transportType) {
        if (transportType.equals("1") || transportType.equals("2")) {
            return 7L;  //立方米
        } else if (transportType.equals("3") || transportType.equals("4")) {
            return 6L;  //千克
        }
        return 0L;
    }

    private void clearAndSavePriceStepList(List<Long> productPriceIdList,  List<ProductPriceStepDO> stepList, Map<Long, String> priceTransportTypeMap, boolean batchUpdate) {
        if (stepList == null) stepList = new ArrayList<>();

        stepList = stepList.stream()
                .filter(t -> t.getStartNum() != null &&
                        t.getWeightUnit() != null &&
                        t.getEndNum() != null)
                .collect(Collectors.toList());

        if (CollectionUtil.isEmpty(stepList)) return;


        Collection<Integer> priceTypes = stepList.stream()
                .map(ProductPriceStepDO::getPriceType)
                .collect(Collectors.toSet());
        priceTypes.add(PriceTypeEnum.FULL_PRICE.getPriceType());
        List<ProductPriceStepDO> dbStepList = productPriceStepMapper.selectList(new LambdaQueryWrapper<ProductPriceStepDO>()
            .in(ProductPriceStepDO::getProductPriceId, productPriceIdList)
                .in(ProductPriceStepDO::getPriceType, priceTypes)
        );
        Map<Long, List<ProductPriceStepDO>> dbStepMap = dbStepList.stream()
                .collect(Collectors.groupingBy(ProductPriceStepDO::getProductPriceId));

        List<ProductPriceStepDO> insertList1 = new ArrayList<>();
        List<ProductPriceStepDO> updateList = new ArrayList<>();
        for (Long productPriceId : productPriceIdList) {
            String transportType = priceTransportTypeMap.get(productPriceId);
            List<ProductPriceStepDO> dbPriceStepList = dbStepMap.get(productPriceId);
            if(CollectionUtil.isEmpty(dbPriceStepList)) {
                for (ProductPriceStepDO stepItem : stepList) {
                    if(stepItem.getTransportPrice() == null && stepItem.getClearancePrice() == null
                        && stepItem.getAllPrice() == null) continue;
                    ProductPriceStepDO dbItem = ObjectUtil.clone(stepItem);
                    dbItem.setId(null);
                    dbItem.setProductPriceId(productPriceId);
                    initBasePrice(dbItem, transportType);
                    insertList1.add(dbItem);
                }
            } else {
                //更新
                for (ProductPriceStepDO stepItem : stepList) {
                    boolean update = false;
                    for (ProductPriceStepDO dbStepItem : dbPriceStepList) {
                        if(!stepItem.getPriceType().equals(dbStepItem.getPriceType())) continue;
                        if(stepItem.getStartNum().intValue() != dbStepItem.getStartNum().intValue() ||
                                stepItem.getEndNum().intValue() != dbStepItem.getEndNum().intValue() ||
                                stepItem.getWeightUnit().longValue() != dbStepItem.getWeightUnit().longValue()) {
                            continue;
                        }
                        //更新
                        update = true;
                        dbStepItem.setSpecialList(stepItem.getSpecialList());
                        dbStepItem.setPackagingList(stepItem.getPackagingList());
                        updateList.add(dbStepItem);
                        if(stepItem.getTransportPrice() == null && stepItem.getClearancePrice() == null
                                && stepItem.getAllPrice() == null) continue;
                        if(stepItem.getTransportPrice() != null) {
                            dbStepItem.setTransportPrice(stepItem.getTransportPrice());
                        }
                        if(stepItem.getTransportPriceUnit() != null) {
                            dbStepItem.setTransportPriceUnit(stepItem.getTransportPriceUnit());
                        }
                        if(stepItem.getTransportVolumeUnit() != null) {
                            dbStepItem.setTransportVolumeUnit(stepItem.getTransportVolumeUnit());
                        }
                        if(stepItem.getClearancePrice() != null) {
                            dbStepItem.setClearancePrice(stepItem.getClearancePrice());
                        }
                        if(stepItem.getClearancePriceUnit() != null) {
                            dbStepItem.setClearancePriceUnit(stepItem.getClearancePriceUnit());
                        }
                        if(stepItem.getClearanceVolumeUnit() != null) {
                            dbStepItem.setClearanceVolumeUnit(stepItem.getClearanceVolumeUnit());
                        }
                        if(stepItem.getAllPrice() != null) {
                            dbStepItem.setAllPrice(stepItem.getAllPrice());
                        }
                        if(stepItem.getAllPriceUnit() != null) {
                            dbStepItem.setAllPriceUnit(stepItem.getAllPriceUnit());
                        }
                        if(stepItem.getAllVolumeUnit() != null) {
                            dbStepItem.setAllVolumeUnit(stepItem.getAllVolumeUnit());
                        }
                        if(stepItem.getPriceType() != null) {
                            dbStepItem.setPriceType(stepItem.getPriceType());
                        }
                        if(stepItem.getRankNum() != null) {
                            dbStepItem.setRankNum(stepItem.getRankNum());
                        }
                        break;
                    }
                    if(!update) {
                        ProductPriceStepDO dbItem = ObjectUtil.clone(stepItem);
                        dbItem.setId(null);
                        dbItem.setProductPriceId(productPriceId);
                        initBasePrice(dbItem, transportType);
                        insertList1.add(dbItem);
                    }
                }
            }
        }
        if(CollectionUtil.isNotEmpty(insertList1)) {
            productPriceStepMapper.insertBatchSomeColumn(insertList1);
        }
        if(CollectionUtil.isNotEmpty(updateList)) {
            productPriceStepMapper.updateBatch(updateList);
        }

        List<Long> dbStepIdList = dbStepList.stream()
                .map(ProductPriceStepDO::getId)
                .collect(toList());
        if(!batchUpdate) {
            List<Long> updateIdList = updateList.stream()
                    .map(ProductPriceStepDO::getId)
                    .collect(toList());
            List<Long> dbIdList = new ArrayList<>(dbStepIdList);
            dbIdList.removeAll(updateIdList);
            if(CollectionUtil.isNotEmpty(dbIdList)) {
                productPriceStepMapper.physicsDeleteById(Joiner.on(",").join(dbIdList));
            }
        }

        insertList1.addAll(updateList);
        List<ProductPriceStepSpecialDO> allUpdateStepSpecialList = insertList1.stream()
                .filter(step -> CollectionUtil.isNotEmpty(step.getSpecialList()))
                .flatMap(step -> step.getSpecialList().stream())
                .filter(stepSpecialDO -> stepSpecialDO.getTransportPrice() != null
                        || stepSpecialDO.getClearancePrice() != null
                        || stepSpecialDO.getAllPrice() != null)
                .collect(Collectors.toList());
        Map<Long, List<ProductPriceStepSpecialDO>> stepSpecialGroupMap = new HashMap<>();
        if(CollectionUtil.isNotEmpty(allUpdateStepSpecialList)) {
            //数据库查找所有阶梯特需数据
            List<List<Long>> stepIdPartList = Lists.partition(dbStepIdList, 1000);
            List<ProductPriceStepSpecialDO> allStepSpecialList = new ArrayList<>();
            for (List<Long> stepIdLis : stepIdPartList) {
                List<ProductPriceStepSpecialDO> itemStepSpecialList = productPriceStepSpecialMapper.selectList(new LambdaQueryWrapper<ProductPriceStepSpecialDO>()
                        .in(ProductPriceStepSpecialDO::getProductPriceStepId, stepIdLis)
                );
                allStepSpecialList.addAll(itemStepSpecialList);
            }
            stepSpecialGroupMap = allStepSpecialList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepSpecialDO::getProductPriceStepId));
        }

        List<ProductPriceStepSpecialDO> specialInsertList = new ArrayList<>();
        List<ProductPriceStepSpecialDO> specialUpdateList = new ArrayList<>();

        for (ProductPriceStepDO stepDO : insertList1) {
            String transportType = priceTransportTypeMap.get(stepDO.getProductPriceId());
            List<ProductPriceStepSpecialDO> productPriceStepSpecialDOS = stepDO.getSpecialList();
            //过滤掉价格为空的数据
            if(CollectionUtil.isNotEmpty(productPriceStepSpecialDOS)) {
                productPriceStepSpecialDOS = productPriceStepSpecialDOS.stream()
                        .filter(stepSpecialDO -> stepSpecialDO.getTransportPrice() != null
                                || stepSpecialDO.getClearancePrice() != null
                                || stepSpecialDO.getAllPrice() != null)
                        .collect(toList());

                List<ProductPriceStepSpecialDO> dbStepSpecialList = stepSpecialGroupMap.get(stepDO.getId());
                Map<String, ProductPriceStepSpecialDO> stepSpecialMap = new HashMap<>();
                if(CollectionUtil.isNotEmpty(dbStepSpecialList)) {
                    stepSpecialMap = dbStepSpecialList.stream()
                            .collect(Collectors.toMap(ProductPriceStepSpecialDO::getSpecialDictType, v -> v));
                }

                for (ProductPriceStepSpecialDO stepSpecialDO : productPriceStepSpecialDOS) {
                    ProductPriceStepSpecialDO dbStepSpecialDO = stepSpecialMap.get(stepSpecialDO.getSpecialDictType());
                    if(dbStepSpecialDO == null) {
                        ProductPriceStepSpecialDO dbItem = ObjectUtil.clone(stepSpecialDO);
                        dbItem.setId(null);
                        dbItem.setProductPriceId(stepDO.getProductPriceId());
                        dbItem.setProductPriceStepId(stepDO.getId());
                        initBasePrice(dbItem, transportType);
                        specialInsertList.add(dbItem);
                    } else {
                        if(stepSpecialDO.getTransportPrice() != null) {
                            dbStepSpecialDO.setTransportPrice(stepSpecialDO.getTransportPrice());
                        }
                        if(stepSpecialDO.getTransportPriceUnit() != null) {
                            dbStepSpecialDO.setTransportPriceUnit(stepSpecialDO.getTransportPriceUnit());
                        }
                        if(stepSpecialDO.getTransportVolumeUnit() != null) {
                            dbStepSpecialDO.setTransportVolumeUnit(stepSpecialDO.getTransportVolumeUnit());
                        }
                        if(stepSpecialDO.getClearancePrice() != null) {
                            dbStepSpecialDO.setClearancePrice(stepSpecialDO.getClearancePrice());
                        }
                        if(stepSpecialDO.getClearancePriceUnit() != null) {
                            dbStepSpecialDO.setClearancePriceUnit(stepSpecialDO.getClearancePriceUnit());
                        }
                        if(stepSpecialDO.getClearanceVolumeUnit() != null) {
                            dbStepSpecialDO.setClearanceVolumeUnit(stepSpecialDO.getClearanceVolumeUnit());
                        }
                        if(stepSpecialDO.getAllPrice() != null) {
                            dbStepSpecialDO.setAllPrice(stepSpecialDO.getAllPrice());
                        }
                        if(stepSpecialDO.getAllPriceUnit() != null) {
                            dbStepSpecialDO.setAllPriceUnit(stepSpecialDO.getAllPriceUnit());
                        }
                        if(stepSpecialDO.getAllVolumeUnit() != null) {
                            dbStepSpecialDO.setAllVolumeUnit(stepSpecialDO.getAllVolumeUnit());
                        }
                        specialUpdateList.add(dbStepSpecialDO);
                    }
                }
            }
        }

        if(CollectionUtil.isNotEmpty(specialInsertList)) {
            productPriceStepSpecialMapper.insertBatchSomeColumn(specialInsertList);
        }
        if(CollectionUtil.isNotEmpty(specialUpdateList)) {
            productPriceStepSpecialMapper.updateBatch(specialUpdateList);
        }

        List<ProductPriceStepPackagingDO> stepPackagingList = insertList1.stream()
                .map(ProductPriceStepDO::getPackagingList)
                .filter(CollectionUtil::isNotEmpty)
                .flatMap(List::stream)
                .collect(Collectors.toList());
        Map<Long, List<ProductPriceStepPackagingDO>> stepPackagingGroupMap = new HashMap<>();
        List<Long> dbPackingIdList = new ArrayList<>();
        if(CollectionUtil.isNotEmpty(stepPackagingList)) {
            //数据库查找所有阶梯包装数据
            List<List<Long>> stepIdPartList = Lists.partition(dbStepIdList, 1000);
            List<ProductPriceStepPackagingDO> allStepPackagingList = new ArrayList<>();
            for (List<Long> stepIdLis : stepIdPartList) {
                List<ProductPriceStepPackagingDO> itemStepPackagingList = productPriceStepPackagingMapper.selectList(new LambdaQueryWrapper<ProductPriceStepPackagingDO>()
                        .in(ProductPriceStepPackagingDO::getProductPriceStepId, stepIdLis)
                );
                allStepPackagingList.addAll(itemStepPackagingList);
            }
            dbPackingIdList = allStepPackagingList.stream()
                    .map(BasePackagingPrice::getId)
                    .collect(toList());
            stepPackagingGroupMap = allStepPackagingList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepPackagingDO::getProductPriceStepId));
        }

        List<ProductPriceStepPackagingDO> packagingInsertList = new ArrayList<>();
        List<ProductPriceStepPackagingDO> packagingUpdateList = new ArrayList<>();
        List<ProductPriceStepPackagingDO> packagingDeleteListList = new ArrayList<>();

        for (ProductPriceStepDO priceStepDO : insertList1) {
            List<ProductPriceStepPackagingDO> productPriceStepPackagingDOS = priceStepDO.getPackagingList();
            if(CollectionUtil.isNotEmpty(productPriceStepPackagingDOS)) {
                productPriceStepPackagingDOS = productPriceStepPackagingDOS.stream()
                        .filter(t -> t.getPackagingPrice() != null)
                        .collect(toList());
            }
            if(CollectionUtil.isNotEmpty(productPriceStepPackagingDOS)) {
                List<ProductPriceStepPackagingDO> dbStepPackagingList = stepPackagingGroupMap.get(priceStepDO.getId());
                if(CollectionUtil.isEmpty(dbStepPackagingList)) {
                    for (int i = 0; i < productPriceStepPackagingDOS.size(); i++) {
                        ProductPriceStepPackagingDO m = productPriceStepPackagingDOS.get(i);
                        ProductPriceStepPackagingDO dbItem = ObjectUtil.clone(m);
                        dbItem.setId(null);
                        dbItem.setProductPriceId(priceStepDO.getProductPriceId());
                        dbItem.setProductPriceStepId(priceStepDO.getId());
                        dbItem.setRankNum(i + 1);
                        dbItem.setDeleted(false);
                        packagingInsertList.add(dbItem);
                    }
                } else {
                    for (int i = 0; i < productPriceStepPackagingDOS.size(); i++) {
                        ProductPriceStepPackagingDO item = productPriceStepPackagingDOS.get(i);
                        String addPricePackagingTypes = item.getPackagingTypes();
                        if(StringUtils.isBlank(addPricePackagingTypes)) continue;
                        Long addPricePackagingPriceUnit = item.getPackagingPriceUnit();
                        if(addPricePackagingPriceUnit == null) continue;
                        Long addPricePackagingVolumeUnit = item.getPackagingVolumeUnit();
                        if(addPricePackagingVolumeUnit == null) continue;
                        Set<String> addPriceTypeSets = new HashSet<>(Arrays.asList(addPricePackagingTypes.split(",")));

                        //新增BC
                        ProductPriceStepPackagingDO dbItem = ObjectUtil.clone(item);
                        dbItem.setId(null);
                        dbItem.setProductPriceId(priceStepDO.getProductPriceId());
                        dbItem.setProductPriceStepId(priceStepDO.getId());
                        dbItem.setRankNum(i + 1);
                        dbItem.setDeleted(false);
                        packagingInsertList.add(dbItem);

                        if(!batchUpdate) continue;
                        for (ProductPriceStepPackagingDO dbStepPacking : dbStepPackagingList) {
                            Set<String> dbPriceTypeSets = new HashSet<>(Arrays.asList(dbStepPacking.getPackagingTypes().split(",")));
                            dbPriceTypeSets.removeAll(addPriceTypeSets);
                            if(dbPriceTypeSets.isEmpty()) {
                                packagingDeleteListList.add(dbStepPacking);
                            } else {
                                dbStepPacking.setPackagingTypes(Joiner.on(",").join(dbPriceTypeSets));
                                dbStepPacking.setPackagingPriceUnit(item.getPackagingPriceUnit());
                                dbStepPacking.setPackagingVolumeUnit(item.getPackagingVolumeUnit());
                                packagingUpdateList.add(dbStepPacking);
                            }
                        }
                    }
                }
            }
        }

        if(CollectionUtil.isNotEmpty(packagingInsertList)) {
            productPriceStepPackagingMapper.insertBatchSomeColumn(packagingInsertList);
        }
        if(CollectionUtil.isNotEmpty(packagingUpdateList)) {
            productPriceStepPackagingMapper.updateBatch(packagingUpdateList);
        }
        if(CollectionUtil.isNotEmpty(packagingDeleteListList)) {
            List<Long> dIdList = packagingDeleteListList.stream()
                    .map(BasePackagingPrice::getId)
                    .collect(toList());
            productPriceStepPackagingMapper.deleteBatchIds(dIdList);
        }
        if(!batchUpdate) {
            List<Long> updateIdList = packagingUpdateList.stream()
                    .map(ProductPriceStepPackagingDO::getId)
                    .collect(toList());
            dbPackingIdList.removeAll(updateIdList);
            if(CollectionUtil.isNotEmpty(dbPackingIdList)) {
                productPriceStepPackagingMapper.physicsDeleteById(Joiner.on(",").join(dbPackingIdList));
            }
        }
    }

    private void initBasePackagingPrice(BasePackagingPrice basePackagingPrice, String transportType) {
        basePackagingPrice.setDeleted(false);
        if(basePackagingPrice.getPackagingPrice() == null) {
            basePackagingPrice.setPackagingPrice(BigDecimal.ZERO);
        }
        if(basePackagingPrice.getPackagingPriceUnit() == null) {
            basePackagingPrice.setPackagingPriceUnit(1L);
        }
        if (basePackagingPrice.getPackagingVolumeUnit() == null) {
            basePackagingPrice.setPackagingVolumeUnit(getUnitByTransportType(transportType));
        }
    }

    private void initBasePrice(BasePrice basePrice, String transportType) {
        if (basePrice.getTransportPrice() == null) {
            basePrice.setTransportPrice(BigDecimal.ZERO);
        }
        if (basePrice.getTransportPriceUnit() == null) {
            basePrice.setTransportPriceUnit(1L);
        }
        if (basePrice.getClearancePrice() == null) {
            basePrice.setClearancePrice(BigDecimal.ZERO);
        }
        if (basePrice.getClearancePriceUnit() == null) {
            basePrice.setClearancePriceUnit(1L);
        }
        if (basePrice.getAllPrice() == null) {
            basePrice.setAllPrice(BigDecimal.ZERO);
        }
        if (basePrice.getAllPriceUnit() == null) {
            basePrice.setAllPriceUnit(1L);
        }

        if (basePrice.getTransportVolumeUnit() == null) {
            basePrice.setTransportVolumeUnit(getUnitByTransportType(transportType));
        }
        if (basePrice.getClearanceVolumeUnit() == null) {
            basePrice.setClearanceVolumeUnit(getUnitByTransportType(transportType));
        }
        if (basePrice.getAllVolumeUnit() == null) {
            basePrice.setAllVolumeUnit(getUnitByTransportType(transportType));
        }
        basePrice.setDeleted(false);
    }

    private void clearAndSaveSpecialList(List<Long> productPriceIdList, List<ProductPriceSpecialDO> specialList, Map<Long, String> priceTransportTypeMap) {
        if (specialList == null) specialList = new ArrayList<>();

        specialList = specialList.stream()
                .filter(t -> t.getTransportPrice() != null ||
                        t.getClearancePrice() != null ||
                        t.getAllPrice() != null)
                .collect(Collectors.toList());

        if (CollectionUtil.isEmpty(specialList)) return;

        List<ProductPriceSpecialDO> insertList = new ArrayList<>();
        List<ProductPriceSpecialDO> updateList = new ArrayList<>();

        List<ProductPriceSpecialDO> specialDOList = productPriceSpecialMapper.selectList(new LambdaQueryWrapper<ProductPriceSpecialDO>()
            .in(ProductPriceSpecialDO::getProductPriceId, productPriceIdList)
        );
        Map<Long, List<ProductPriceSpecialDO>> specialMap = specialDOList.stream()
                .collect(Collectors.groupingBy(ProductPriceSpecialDO::getProductPriceId));

//        productPriceSpecialMapper.physicsDelete(Joiner.on(",").join(productPriceIdList));

        //首先删除所有的特殊需求列表
        for (Long productPriceId : productPriceIdList) {
            List<ProductPriceSpecialDO> productPriceSpecialDOList = specialMap.get(productPriceId);
            String transportType = priceTransportTypeMap.get(productPriceId);
            if(CollectionUtil.isEmpty(productPriceSpecialDOList)) {
                specialList.forEach(t -> {
                    ProductPriceSpecialDO dbItem = ObjectUtil.clone(t);
                    dbItem.setId(null);
                    dbItem.setProductPriceId(productPriceId);
                    initBasePrice(dbItem, transportType);
                    insertList.add(dbItem);
                });
            } else {
                Map<String, ProductPriceSpecialDO> dictTypeMap = productPriceSpecialDOList.stream()
                        .collect(Collectors.toMap(ProductPriceSpecialDO::getSpecialDictType, v -> v));
                specialList.forEach(t -> {
                    String dictType = t.getSpecialDictType();
                    ProductPriceSpecialDO productPriceSpecialDO = dictTypeMap.get(dictType);
                    if(productPriceSpecialDO == null) {
                        ProductPriceSpecialDO dbItem = ObjectUtil.clone(t);
                        dbItem.setId(null);
                        dbItem.setProductPriceId(productPriceId);
                        initBasePrice(dbItem, transportType);
                        insertList.add(dbItem);
                    } else {
                        //更新
                        updateSpecialColumn(productPriceSpecialDO, t);
                        updateList.add(productPriceSpecialDO);
                    }
                });
            }

        }
        if(CollectionUtil.isNotEmpty(insertList)) {
            productPriceSpecialMapper.insertBatchSomeColumn(insertList);
        }

        if(CollectionUtil.isNotEmpty(updateList)) {
            productPriceSpecialMapper.updateBatch(updateList);
        }
        List<Long> dbIdList = specialDOList.stream()
                .map(ProductPriceSpecialDO::getId)
                .collect(toList());
        List<Long> updateIdList = updateList.stream()
                .map(ProductPriceSpecialDO::getId)
                .collect(toList());
        dbIdList.removeAll(updateIdList);
        if(CollectionUtil.isNotEmpty(dbIdList)) {
//            productPriceSpecialMapper.physicsDeleteById(Joiner.on(",").join(dbIdList));
        }

    }

    private void updateSpecialColumn(ProductPriceSpecialDO productPriceSpecialDO, ProductPriceSpecialDO updateDO) {
        if (Objects.nonNull(updateDO.getTransportPrice())) {
            productPriceSpecialDO.setTransportPrice(updateDO.getTransportPrice());
            productPriceSpecialDO.setTransportPriceUnit(updateDO.getTransportPriceUnit());
            productPriceSpecialDO.setTransportVolumeUnit(updateDO.getTransportVolumeUnit());
        }
        if (Objects.nonNull(updateDO.getClearancePrice())) {
            productPriceSpecialDO.setClearancePrice(updateDO.getClearancePrice());
            productPriceSpecialDO.setClearancePriceUnit(updateDO.getClearancePriceUnit());
            productPriceSpecialDO.setClearanceVolumeUnit(updateDO.getClearanceVolumeUnit());
        }

        if (Objects.nonNull(updateDO.getAllPrice())) {
            productPriceSpecialDO.setAllPrice(updateDO.getAllPrice());
            productPriceSpecialDO.setAllPriceUnit(updateDO.getAllPriceUnit());
            productPriceSpecialDO.setAllVolumeUnit(updateDO.getAllVolumeUnit());
        }
    }

    private void clearAndSavePackagingList(List<Long> productPriceIdList, List<ProductPricePackagingDO> packagingDOList, Map<Long, String> priceTransportTypeMap, boolean batchUpdate) {
        if (packagingDOList == null) packagingDOList = new ArrayList<>();

        packagingDOList = packagingDOList.stream()
                .filter(t -> t.getPackagingPrice() != null)
                .collect(Collectors.toList());

        if (CollectionUtil.isEmpty(packagingDOList)) return;

        List<ProductPricePackagingDO> insertList = new ArrayList<>();
        List<ProductPricePackagingDO> updateList = new ArrayList<>();
        List<ProductPricePackagingDO> deleteList = new ArrayList<>();

        Collection<Integer> priceTypes = packagingDOList.stream()
                .map(ProductPricePackagingDO::getPriceType)
                .collect(Collectors.toSet());
        priceTypes.add(PriceTypeEnum.FULL_PRICE.getPriceType());
        List<ProductPricePackagingDO> dbPackagingList = productPricePackagingMapper.selectList(new LambdaQueryWrapper<ProductPricePackagingDO>()
                .in(ProductPricePackagingDO::getProductPriceId, productPriceIdList)
                .in(ProductPricePackagingDO::getPriceType, priceTypes)
        );
        Map<Long, List<ProductPricePackagingDO>> dbPackagingMap = dbPackagingList.stream()
                .collect(Collectors.groupingBy(ProductPricePackagingDO::getProductPriceId));

        //首先删除所有的包装价格列表
        for (Long productPriceId : productPriceIdList) {
            String transportType = priceTransportTypeMap.get(productPriceId);
            List<ProductPricePackagingDO> dbPackagingDOList = dbPackagingMap.get(productPriceId);
            //没有历史数据则新增
            if(CollectionUtil.isEmpty(dbPackagingDOList)) {
                for (int i = 0; i < packagingDOList.size(); i++) {
                    ProductPricePackagingDO t = packagingDOList.get(i);
                    ProductPricePackagingDO dbItem = ObjectUtil.clone(t);
                    dbItem.setId(null);
                    dbItem.setRankNum(i + 1);
                    dbItem.setProductPriceId(productPriceId);
                    initBasePackagingPrice(dbItem, transportType);
                    insertList.add(dbItem);
                }
            } else {
                //更新
                int i = 0;
                for (ProductPricePackagingDO item : packagingDOList) {
                    String addPricePackagingTypes = item.getPackagingTypes();
                    if(StringUtils.isBlank(addPricePackagingTypes)) continue;
                    Long addPricePackagingPriceUnit = item.getPackagingPriceUnit();
                    if(addPricePackagingPriceUnit == null) continue;
                    Long addPricePackagingVolumeUnit = item.getPackagingVolumeUnit();
                    if(addPricePackagingVolumeUnit == null) continue;
                    i++;
                    Set<String> addPriceTypeSets = new HashSet<>(Arrays.asList(addPricePackagingTypes.split(",")));

                    ProductPricePackagingDO dbItem = ObjectUtil.clone(item);
                    dbItem.setId(null);
                    dbItem.setRankNum(i + 1);
                    dbItem.setProductPriceId(productPriceId);
                    initBasePackagingPrice(dbItem, transportType);
                    insertList.add(dbItem);

                    if(!batchUpdate) continue;
                    for (ProductPricePackagingDO dbPackagingItem : dbPackagingDOList) {
                        if(!item.getPriceType().equals(dbPackagingItem.getPriceType())) continue;
                        Set<String> dbPriceTypeSets = new HashSet<>(Arrays.asList(dbPackagingItem.getPackagingTypes().split(",")));
                        dbPriceTypeSets.removeAll(addPriceTypeSets);
                        if(dbPriceTypeSets.isEmpty()) {
                            deleteList.add(dbPackagingItem);
                        } else {
                            dbPackagingItem.setPackagingTypes(Joiner.on(",").join(dbPriceTypeSets));
                            dbPackagingItem.setPackagingPriceUnit(item.getPackagingPriceUnit());
                            dbPackagingItem.setPackagingVolumeUnit(item.getPackagingVolumeUnit());
                            if(item.getPriceType() != null) {
                                dbPackagingItem.setPriceType(item.getPriceType());
                            }
                            updateList.add(dbPackagingItem);
                        }
                    }
                }
            }

        }
        if(CollectionUtil.isNotEmpty(insertList)) {
            productPricePackagingMapper.insertBatchSomeColumn(insertList);
        }
        if(CollectionUtil.isNotEmpty(updateList)) {
            productPricePackagingMapper.updateBatch(updateList);
        }
        if(CollectionUtil.isNotEmpty(deleteList)) {
            List<Long> dIdList = deleteList.stream()
                    .map(BasePackagingPrice::getId)
                    .collect(toList());
            productPricePackagingMapper.deleteBatchIds(dIdList);
        }

        if(!batchUpdate) {
            List<Long> dbIdList = dbPackagingList.stream()
                    .map(ProductPricePackagingDO::getId)
                    .collect(toList());
            List<Long> updateIdList = updateList.stream()
                    .map(ProductPricePackagingDO::getId)
                    .collect(toList());
            dbIdList.removeAll(updateIdList);
            if(CollectionUtil.isNotEmpty(dbIdList)) {
                productPricePackagingMapper.physicsDeleteById(Joiner.on(",").join(dbIdList));
            }
        }

    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void updateProductPrice(ProductPriceUpdateReqVO updateReqVO) {
        // 校验存在
        Long priceId = updateReqVO.getId();
        this.validateProductPriceExists(priceId);
        // 更新
        ProductPriceDO updateObj = ProductPriceConvert.INSTANCE.convert(updateReqVO);

        if (null != updateReqVO.getStepPrice() && updateReqVO.getStepPrice() == 1) {
            checkAndUpdateStepToMainPrice(updateObj, updateReqVO);
        }

        ProductPriceDO priceDO = getProductPrice(priceId);
        //添加修改日志
        batchAddSnapshot(Collections.singletonList(priceDO));

        productPriceMapper.updateById(updateObj);
        if (updateObj.getValidateStartDate() == null || updateObj.getValidateEndDate() == null) {
            productPriceMapper.update(null, Wrappers.<ProductPriceDO>lambdaUpdate()
                    .set(ProductPriceDO::getValidateEndDate, updateObj.getValidateEndDate())
                    .set(ProductPriceDO::getValidateStartDate, updateObj.getValidateStartDate())
                    .eq(ProductPriceDO::getId, priceId)
            );
        }


        WarehouseLineDO warehouseLineDO = warehouseLineMapper.selectOne("id", priceDO.getWarehouseLineId());
        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        priceTransportTypeMap.put(priceId, warehouseLineDO.getTransportType());

        //插入特殊属性
        List<ProductPriceSpecialDO> specialList = updateReqVO.getSpecialList();
        clearAndSaveSpecialList(Collections.singletonList(priceId), specialList, priceTransportTypeMap);

        //处理包装单位价格
        List<ProductPricePackagingDO> allPackageList = buildAllPackageList(updateReqVO);
        clearAndSavePackagingList(Collections.singletonList(priceId), allPackageList, priceTransportTypeMap, false);

        //判断是否是阶梯定价
        if (null != updateReqVO.getStepPrice() && updateReqVO.getStepPrice() == 1) {
            List<ProductPriceStepDO> allStepList = buildAllStepList(updateReqVO);
            clearAndSavePriceStepList(Collections.singletonList(updateReqVO.getId()), allStepList, priceTransportTypeMap, false);
        }

        handleOrderException(Collections.singletonList(priceId), true);
    }

    /**
     * 处理订单异常,未设置价格或者价格已经过期的会生成未报价异常
     *
     * @param productPriceIdList 路线价格ID列表
     */
    private void handleOrderException(List<Long> productPriceIdList, boolean sync) {
        List<ProductPriceDO> priceList = getProductPriceList(productPriceIdList);

        List<ProductPriceDO> needHandleList = new ArrayList<>();

        Date now = new Date();

        for (ProductPriceDO productPriceDO : priceList) {
            Date validateStartDate = productPriceDO.getValidateStartDate();
            Date validateEndDate = productPriceDO.getValidateEndDate();

            boolean hasPrice = false;
            if (productPriceDO.getPriceType() == 0) {
                BigDecimal transportPrice = productPriceDO.getTransportPrice();
                hasPrice = transportPrice != null && transportPrice.floatValue() > 0;
            } else if (productPriceDO.getPriceType() == 1) {
                BigDecimal allPrice = productPriceDO.getAllPrice();
                hasPrice = allPrice != null && allPrice.floatValue() > 0;
            }

            if (!hasPrice) continue;

            //
            if (validateStartDate == null && validateEndDate == null) {
                needHandleList.add(productPriceDO);
            } else if (validateStartDate != null && validateEndDate != null) {
                //判断当前时间是否在这个区间
                if (now.compareTo(validateEndDate) < 0 && now.compareTo(validateStartDate) > 0) {
                    needHandleList.add(productPriceDO);
                }
            } else if (validateStartDate != null) {
                if (now.compareTo(validateStartDate) > 0) {
                    needHandleList.add(productPriceDO);
                }
            } else {
                if (now.compareTo(validateEndDate) < 0) {
                    needHandleList.add(productPriceDO);
                }
            }
        }

        if (CollectionUtil.isNotEmpty(needHandleList)) {
            HandleData handleData;
            if (sync) {
                handleData = new HandleExceptionEventSync();
            } else {
                handleData = new HandleExceptionEvent();
            }
            handleData.setNeedHandleList(needHandleList);
            applicationContext.publishEvent(handleData);
        }
    }

    @Override
    public void deleteProductPrice(Long id) {
        // 校验存在
        this.validateProductPriceExists(id);
        // 删除
        productPriceMapper.deleteById(id);
    }

    private void validateProductPriceExists(Long id) {
        if (productPriceMapper.selectById(id) == null) {
            throw exception(PRODUCT_PRICE_NOT_EXISTS);
        }
    }

    @Override
    public ProductPriceDO getProductPrice(Long id) {
        return productPriceMapper.selectById(id);
    }

    @Override
    public List<ProductPriceDO> getProductPriceList(Collection<Long> ids) {
        return productPriceMapper.selectBatchIds(ids);
    }

    @Override
    public void batchOff(Collection<Long> ids) {
        for (Long id : ids) {
            ProductPriceDO productPriceDO = new ProductPriceDO();
            productPriceDO.setId(id);
            productPriceDO.setAuditStatus(1);
            this.updateById(productPriceDO);
        }
    }

    @Override
    public void batchBlock(Collection<Long> ids) {
        for (Long id : ids) {
            ProductPriceDO productPriceDO = new ProductPriceDO();
            productPriceDO.setId(id);
            productPriceDO.setBlacklist(CommonStatusEnum.DISABLE.getStatus());
            this.updateById(productPriceDO);
        }
    }

    @Override
    public PageResult<ProductPriceDO> getProductPricePage(ProductPricePageReqVO pageParam) {
        initDictParams(pageParam);

        IPage<ProductPriceDO> mpPage = MyBatisUtils.buildPage(pageParam);
        mpPage = productPriceMapper.getPage(mpPage, pageParam);
        return new PageResult<>(mpPage.getRecords(), mpPage.getTotal());
    }

    @Override
    public void initDictParams(ProductPricePageReqVO pageParam) {
        List<DictDataRespDTO> dictList = dictDataApi.getDictDatas("transport_type");
        List<String> existTypeList = dictList.stream()
                .map(DictDataRespDTO::getValue)
                .collect(Collectors.toList());
        if (CollectionUtil.isNotEmpty(existTypeList)) {
            pageParam.setExistTypeList(existTypeList);
        }
    }

    @Override
    public List<ProductPriceDO> getProductPriceList(ProductPriceExportReqVO exportReqVO) {
        return productPriceMapper.selectList(exportReqVO);
    }

    @Override
    public ProductPriceCheckAO check(ProductPriceReqBaseVO productPriceReqBaseVO) {
        if(productPriceReqBaseVO.isCheckStepPrice()) {
            checkStepPrice(productPriceReqBaseVO);
        }
        List<LineChannelVo> lineChannelList = productPriceReqBaseVO.getLineChannelList();
        Integer isAllProduct = productPriceReqBaseVO.getIsAllProduct();
        List<Long> productIdList = productPriceReqBaseVO.getProductIdList();
        return checkUpdatingAndArgs(lineChannelList, isAllProduct, productIdList);
    }


    private ProductPriceCheckAO checkUpdatingAndArgs(List<LineChannelVo> lineChannelList, Integer isAllProduct, List<Long> productIdList) {
        if (CollectionUtil.isEmpty(lineChannelList)) {
            throw new ServiceException(500, "lineChannelList不能为空");
        }
        List<ProductDO> productList;
        String content;
        if (isAllProduct == 1) {
            productList = baseMapper.getProductList();
            content = "所有商品修改线路价格成功";
        } else {
            if (CollectionUtil.isEmpty(productIdList)) throw new IllegalArgumentException("productIdList不能为空");
            productList = productService.getProductList(productIdList);
            List<String> nameList = productList.stream()
                    .map(ProductDO::getTitleZh)
                    .collect(Collectors.toList());
            content = "【" + Joiner.on(",").join(nameList) + "】的线路价格已完成，请前往查看";
        }
        List<WarehouseLineDO> warehouseLineList = warehouseService.openedRouterList(new WarehouseLineSearchVO());
        //校验正在更新的商品
        checkUpdatingProduct(productList, warehouseLineList);

        ProductPriceCheckAO productPriceCheckAO = new ProductPriceCheckAO();
        productPriceCheckAO.setContent(content);
        productPriceCheckAO.setProductList(productList);
        productPriceCheckAO.setWarehouseLineList(warehouseLineList);
        return productPriceCheckAO;
    }

    private void checkStepPrice(ProductPriceReqBaseVO productPriceStepBaseVO) {
        checkStepPrice(productPriceStepBaseVO.getFreightPriceStepList(), PriceTypeEnum.FREIGHT);
        checkStepPrice(productPriceStepBaseVO.getClearancePriceStepList(), PriceTypeEnum.CLEARANCE);
        checkStepPrice(productPriceStepBaseVO.getFullPriceStepList(), PriceTypeEnum.FULL_PRICE);
    }

    @Override
    public void batchUpdateProductPrice(ProductPriceBatchUpdateReqVo updateReqVO, ProductPriceCheckAO productPriceCheckAO) {
        String json = JSON.toJSONString(updateReqVO);
        List<ProductDO> productList = productPriceCheckAO.getProductList();
        List<WarehouseLineDO> warehouseLineList = productPriceCheckAO.getWarehouseLineList();
        logAndDelKey(json, productList, updateReqVO.getLineChannelList());

        if (updateReqVO.getBatchMode() == 1) {
            int countBatch = (int) Math.ceil((double) productList.size() / DATA_THRESHOLD);
            List<List<ProductDO>> splitList = Stream.iterate(0, n -> n + 1)
                    .limit(countBatch)
                    .parallel()
                    .map(a -> productList.stream().skip(a * DATA_THRESHOLD).limit(DATA_THRESHOLD).parallel().collect(toList())
                    ).collect(toList());
            for (List<ProductDO> itemList : splitList) {
                ProductPriceCreateReqVO createReqVO = new ProductPriceCreateReqVO();
                BeanUtils.copyProperties(updateReqVO, createReqVO);
                createReqVO.setBatch(true);
                batchUpdate(createReqVO, warehouseLineList, itemList);
            }
        } else {
            //???
            for (ProductDO productDO : productList) {
                ProductPriceCreateReqVO createReqVO = new ProductPriceCreateReqVO();
                BeanUtils.copyProperties(updateReqVO, createReqVO);
                createReqVO.setBatch(true);
                createReqVO.setProductId(productDO.getId());
                createProductPrice(createReqVO, warehouseLineList);
            }
        }
        //发送站内信
        sendMessage(productPriceCheckAO.getContent());
    }

    private void logAndDelKey(String json, List<ProductDO> productList, List<LineChannelVo> lineChannelList) {
        ProductBatchLogDO logDO = new ProductBatchLogDO();
        logDO.setJson(json);
        productBatchLogService.save(logDO);
        //商品线路保存在redis。更新成功后移除
        batchSaveKeyInCache(productList, lineChannelList);
    }

    @Override
    public void batchUpdateProductPriceAir(ProductPriceBatchUpdateAirReqVo updateReqVO, ProductPriceCheckAO productPriceCheckAO) {
        String json = JSON.toJSONString(updateReqVO);
        List<ProductDO> productList = productPriceCheckAO.getProductList();
        List<WarehouseLineDO> warehouseLineList = productPriceCheckAO.getWarehouseLineList();
        logAndDelKey(json, productList, updateReqVO.getLineChannelList());
        if (updateReqVO.getBatchMode() == 1) {
            int countBatch = (int) Math.ceil((double) productList.size() / DATA_THRESHOLD);
            List<List<ProductDO>> splitList = Stream.iterate(0, n -> n + 1)
                    .limit(countBatch)
                    .parallel()
                    .map(a -> productList.stream().skip(a * DATA_THRESHOLD).limit(DATA_THRESHOLD).parallel().collect(toList())
                    ).collect(toList());
            for (List<ProductDO> itemList : splitList) {
                ProductPriceAirCreateReqVO createReqVO = new ProductPriceAirCreateReqVO();
                BeanUtils.copyProperties(updateReqVO, createReqVO);
                createReqVO.setBatch(true);
                batchUpdateAir(createReqVO, warehouseLineList, itemList);
            }
        } else {
            //batchMode???
            for (ProductDO productDO : productList) {
                ProductPriceAirCreateReqVO createReqVO = new ProductPriceAirCreateReqVO();
                BeanUtils.copyProperties(updateReqVO, createReqVO);
                createReqVO.setProductId(productDO.getId());
                createReqVO.setBatch(true);
                batchUpdateAir(createReqVO, warehouseLineList, Collections.singletonList(productDO));
            }
        }
        //发送站内信
        sendMessage(productPriceCheckAO.getContent());
    }

    private void checkUpdatingProduct(List<ProductDO> productList, List<WarehouseLineDO> warehouseLineList) {
        Iterable<String> keyIterable = redissonClient.getKeys().getKeysByPattern(PREFIX + "*");
        Iterator<String> keyIt = keyIterable.iterator();

        Map<Long, String> productNameMap = productList.stream()
                .collect(Collectors.toMap(ProductDO::getId, ProductDO::getTitleZh, (v1, v2) -> v2));
        Map<Long, String> lineNameMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, t -> t.getStartTitleZh() + "-" + t.getDestTitleZh(), (v1, v2) -> v2));

        List<String> dealingList = new ArrayList<>();
        while (keyIt.hasNext()) {
            String key = keyIt.next();
            key = key.replace(PREFIX, "");
            String[] productLineIdArray = key.split("_");
            Long productId = Long.valueOf(productLineIdArray[0]);
            Long lineId = Long.valueOf(productLineIdArray[1]);
            if (productNameMap.containsKey(productId)) {
                String productName = productNameMap.get(productId);
                String lineName = lineNameMap.get(lineId);
                String tipMessage = productName + "(" + lineName + ")";
                dealingList.add(tipMessage);
                if (dealingList.size() > 30) {
                    dealingList.add("......");
                    break;
                }
            }
        }
        if (CollectionUtil.isNotEmpty(dealingList)) {
            String message = Joiner.on(",").join(dealingList);
            throw new ServiceException(500, message + "正在修改中");
        }
    }

    private void batchSaveKeyInCache(List<ProductDO> productList, List<LineChannelVo> lineChannelList) {
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());

        RBuckets buckets = redissonClient.getBuckets();
        Map<String, Integer> map = new HashMap<>();
        for (ProductDO productDO : productList) {
            Long productId = productDO.getId();
            for (Long lineId : lineIdList) {
                String key = PREFIX + productId + "_" + lineId;
                map.put(key, 1);
            }
        }
        buckets.set(map);
        //暂未找到同时设置的
        RBatch batch = redissonClient.createBatch();
        for (Map.Entry<String, Integer> entry : map.entrySet()) {
            String key = entry.getKey();
            RBucketAsync bucket = batch.getBucket(key);
            Duration duration = Duration.ofHours(2);
            bucket.expireAsync(duration);
//            batch.getBucket(key).expire(2, TimeUnit.HOURS);
        }
        batch.execute();
        System.out.println("执行完毕z");
    }

    private void sendMessage(String content) {
        Long loginUserId = SecurityFrameworkUtils.getLoginUserId();
        InternalMessageCreateDto dto = InternalMessageCreateDto.builder().toIdList(Collections.singletonList(loginUserId))
                .fromId("")
                .sendTime(new Date())
                .title("批量修改线路价格成功")
                .content(content)
                .type(3).build();
        internalMessageApi.createInternalMessage(dto);
    }

    @Override
    public ProductPriceDO getProductPriceById(Long productId, Long lineId, Long shippingChannelId) {
        return productPriceMapper.selectOne(new LambdaQueryWrapperX<ProductPriceDO>()
                .eq(ProductPriceDO::getWarehouseLineId, lineId)
                .eqIfPresent(ProductPriceDO::getShippingChannelId, shippingChannelId)
                .eq(ProductPriceDO::getProductId, productId));
    }

    @Override
    public ProductPriceRespVO getProductPriceDetail(Long id) {
        ProductPriceDO productPrice = getProductPrice(id);
        ProductPriceRespVO productPriceRespVO = ProductPriceConvert.INSTANCE.convert(productPrice);
        //设置线路实体
        Long lineId = productPrice.getWarehouseLineId();
        List<WarehouseLineDO> lineDOList = warehouseService.selectBylineIdList(Arrays.asList(lineId));
        productPriceRespVO.setWarehouseLineDO(lineDOList.get(0));

        //设置商品实体
        Long productId = productPrice.getProductId();
        productPriceRespVO.setProductDO(productService.getProduct(productId));

        //设置特殊需求列表
        productPriceRespVO.setSpecialList(productPriceSpecialService.getProductPriceSpecialList(id));

        List<ProductPriceStepDO> priceStepDOList = productPriceStepService.getProductPriceStepList(id);

        List<Long> stepIdList = priceStepDOList.stream()
                .map(ProductPriceStepDO::getId)
                .collect(toList());
        if(CollectionUtil.isNotEmpty(stepIdList)) {
            List<ProductPriceStepSpecialDO> allStepSpecialList = productPriceStepSpecialMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepSpecialDO>()
                    .in(ProductPriceStepSpecialDO::getProductPriceStepId, stepIdList)
            );
            List<ProductPriceStepPackagingDO> allStepPackagingList = productPriceStepPackagingMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepPackagingDO>()
                    .in(ProductPriceStepPackagingDO::getProductPriceStepId, stepIdList)
            );
            matchProductPriceStep(priceStepDOList, allStepSpecialList, allStepPackagingList);
        }

        //设置包装费用
        List<ProductPricePackagingDO> allPackagingList = productPricePackagingMapper.selectList(new LambdaQueryWrapperX<ProductPricePackagingDO>()
                .eq(ProductPricePackagingDO::getProductPriceId, id)
        );
        buildPackagingTypeList(allPackagingList, productPriceRespVO);

        productPriceRespVO.setCreatorName(adminUserApi.getUserName(productPriceRespVO.getCreator()));
        productPriceRespVO.setUpdaterName(adminUserApi.getUserName(productPriceRespVO.getUpdater()));

        buildStepTypeList(priceStepDOList, productPriceRespVO);
        //设置创建和更新人
        return productPriceRespVO;
    }

    private void matchStepSpecialAndPackaging(List<ProductPriceRespVO> list) {
        List<Long> priceIdList = list.stream()
                .map(ProductPriceRespVO::getId)
                .collect(toList());
        if(CollectionUtil.isEmpty(priceIdList)) return;
        List<ProductPriceStepDO> productPriceStepDOList = productPriceStepService.list(new LambdaQueryWrapper<ProductPriceStepDO>()
                .in(ProductPriceStepDO::getProductPriceId, priceIdList)
        );
        if(CollectionUtil.isNotEmpty(productPriceStepDOList)) {
            Map<Long, List<ProductPriceStepDO>> priceStepMap = productPriceStepDOList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepDO::getProductPriceId));

            List<Long> stepIdList = productPriceStepDOList.stream()
                    .map(ProductPriceStepDO::getId)
                    .collect(toList());
            List<ProductPriceStepSpecialDO> allStepSpecialList = productPriceStepSpecialMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepSpecialDO>()
                    .in(ProductPriceStepSpecialDO::getProductPriceStepId, stepIdList)
            );

            List<ProductPriceStepPackagingDO> allStepPackagingList = productPriceStepPackagingMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepPackagingDO>()
                    .in(ProductPriceStepPackagingDO::getProductPriceStepId, stepIdList)
            );

            list.forEach(t -> {
                List<ProductPriceStepDO> priceStepList = priceStepMap.get(t.getId());
                matchProductPriceStep(priceStepList, allStepSpecialList, allStepPackagingList);
                buildStepTypeList(priceStepList, t);
            });
        }

    }

    private void buildStepTypeList(List<ProductPriceStepDO> priceStepDOList, ProductPriceRespVO productPriceRespVO) {
        if(CollectionUtil.isNotEmpty(priceStepDOList)) {
            Map<Integer, List<ProductPriceStepDO>> priceTypeStepMap = priceStepDOList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepDO::getPriceType));
            for (Map.Entry<Integer, List<ProductPriceStepDO>> entry : priceTypeStepMap.entrySet()) {
                Integer priceType = entry.getKey();
                List<ProductPriceStepDO> priceTypeStepList = entry.getValue();
                priceTypeStepList.sort(Comparator.comparing(ProductPriceStepDO::getRankNum,Comparator.nullsLast(Comparator.naturalOrder())));
                if(priceType == PriceTypeEnum.FREIGHT.getPriceType()) {
                    productPriceRespVO.setFreightPriceStepList(priceTypeStepList);
                } else if(priceType == PriceTypeEnum.FULL_PRICE.getPriceType()) {
                    productPriceRespVO.setFullPriceStepList(priceTypeStepList);
                } else if(priceType == PriceTypeEnum.CLEARANCE.getPriceType()) {
                    productPriceRespVO.setClearancePriceStepList(priceTypeStepList);
                }
            }
        }
    }

    private void buildPackagingTypeList(List<ProductPricePackagingDO> pricePackagingDOList, ProductPriceRespVO productPriceRespVO) {
        if(CollectionUtil.isNotEmpty(pricePackagingDOList)) {
            Map<Integer, List<ProductPricePackagingDO>> priceTypePackagingMap = pricePackagingDOList.stream()
                    .collect(Collectors.groupingBy(ProductPricePackagingDO::getPriceType));
            for (Map.Entry<Integer, List<ProductPricePackagingDO>> entry : priceTypePackagingMap.entrySet()) {
                Integer priceType = entry.getKey();
                List<ProductPricePackagingDO> priceTypePackagingList = entry.getValue();
                priceTypePackagingList.sort(Comparator.comparing(ProductPricePackagingDO::getRankNum));
                if(priceType == PriceTypeEnum.FREIGHT.getPriceType()) {
                    productPriceRespVO.setFreightPricePackagingList(priceTypePackagingList);
                } else if(priceType == PriceTypeEnum.FULL_PRICE.getPriceType()) {
                    productPriceRespVO.setFullPricePackagingList(priceTypePackagingList);
                } else if(priceType == PriceTypeEnum.CLEARANCE.getPriceType()) {
                    productPriceRespVO.setClearancePricePackagingList(priceTypePackagingList);
                }
            }
        }
    }

    private void matchProductPriceStep(List<ProductPriceStepDO> priceStepDOList, List<ProductPriceStepSpecialDO> allStepSpecialList, List<ProductPriceStepPackagingDO> allStepPackagingList) {
        if(CollectionUtil.isNotEmpty(priceStepDOList)) {
            Map<Long, List<ProductPriceStepSpecialDO>> stepSpecialMap = allStepSpecialList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepSpecialDO::getProductPriceStepId));

            Map<Long, List<ProductPriceStepPackagingDO>> stepPackagingMap = allStepPackagingList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepPackagingDO::getProductPriceStepId));

            for (ProductPriceStepDO productPriceStepDO : priceStepDOList) {
                productPriceStepDO.setSpecialList(stepSpecialMap.get(productPriceStepDO.getId()));

                List<ProductPriceStepPackagingDO> stepPackagingList = stepPackagingMap.get(productPriceStepDO.getId());
                if(CollectionUtil.isNotEmpty(stepPackagingList)) {
                    stepPackagingList.sort(Comparator.comparing(BasePackagingPrice::getRankNum));
                    productPriceStepDO.setPackagingList(stepPackagingList);
                }
            }
        }

    }

    @Override
    public ProductPriceRespVO getProductPriceDetail(PriceParam priceParam) {
        Long productId = priceParam.getProductId();
        Long lineId = priceParam.getLineId();
        Long shippingChannelId = priceParam.getChannelId();
        ProductPriceDO dbItem = productPriceMapper.selectOne(new LambdaQueryWrapperX<ProductPriceDO>()
                .eq(ProductPriceDO::getWarehouseLineId, lineId)
                .eqIfPresent(ProductPriceDO::getShippingChannelId, shippingChannelId)
                .eq(ProductPriceDO::getProductId, productId));
        if (dbItem != null) {
            return getProductPriceDetail(dbItem.getId());
        }
        return null;
    }

    @Override
    public void matchLineAndProduct(List<ProductPriceRespVO> list) {
        if (CollectionUtil.isNotEmpty(list)) {
            List<Long> lineIdList = list.stream()
                    .map(ProductPriceBaseVO::getWarehouseLineId)
                    .collect(Collectors.toList());
            List<WarehouseLineDO> lineDOList = warehouseService.selectBylineIdList(lineIdList);

            List<Long> productIdList = list.stream()
                    .map(ProductPriceBaseVO::getProductId)
                    .collect(Collectors.toList());
            List<ProductDO> productDOList = productService.getProductList(productIdList);
            Map<Long, ProductDO> productMap = productDOList.stream()
                    .collect(Collectors.toMap(ProductDO::getId, (v -> v), (v1, v2) -> v2));

            Map<Long, WarehouseLineDO> lineMap = lineDOList.stream()
                    .collect(Collectors.toMap(WarehouseLineDO::getId, (v -> v), (v1, v2) -> v2));
            list.forEach(t -> {
                t.setWarehouseLineDO(lineMap.get(t.getWarehouseLineId()));
                t.setProductDO(productMap.get(t.getProductId()));
            });
        }
        matchStepSpecialAndPackaging(list);
    }

    @Override
    public long getExcelListCount(ProductPricePageReqVO params) {
        return baseMapper.getExcelListCount(params);
    }

    @Override
    public List<ProductPriceExcelData> getExcelList(int pageSize, Long maxId, ProductPricePageReqVO params) {
        List<ProductPriceExcelData> list = baseMapper.getExcelList(pageSize, maxId, params);
        if (CollectionUtil.isNotEmpty(list)) {
            List<Long> idList = list.stream()
                    .map(ProductPriceExcelData::getId)
                    .collect(Collectors.toList());
            List<ProductPriceSpecialDO> specialList = productPriceSpecialService.getProductPriceSpecialList(idList);
            Map<Long, List<ProductPriceSpecialDO>> specialMap = specialList.stream()
                    .collect(Collectors.groupingBy(ProductPriceSpecialDO::getProductPriceId));


            List<ProductPriceStepDO> stepList = productPriceStepService.getProductPriceStepList(idList);
            List<Long> stepIdList = stepList.stream()
                    .map(ProductPriceStepDO::getId)
                    .collect(toList());
            if(CollectionUtil.isNotEmpty(stepIdList)) {
                List<ProductPriceStepSpecialDO> allStepSpecialList = productPriceStepSpecialMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepSpecialDO>()
                        .in(ProductPriceStepSpecialDO::getProductPriceStepId, stepIdList)
                );
                List<ProductPriceStepPackagingDO> allStepPackagingList = productPriceStepPackagingMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepPackagingDO>()
                        .in(ProductPriceStepPackagingDO::getProductPriceStepId, stepIdList)
                );
                matchProductPriceStep(stepList, allStepSpecialList, allStepPackagingList);
            }
            Map<Long, List<ProductPriceStepDO>> stepMap = stepList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepDO::getProductPriceId));

            for (ProductPriceExcelData priceExcelData : list) {
                Long priceId = priceExcelData.getId();
                priceExcelData.setSpecialList(specialMap.get(priceId));
                priceExcelData.setStepList(stepMap.get(priceId));
            }
        }
        return list;
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void copy() {
        new Thread(new Runnable() {
            @Override
            public void run() {
                //广州到alaba的线路
                List<WarehouseLineDO> sourceList = warehouseLineMapper.selectList(new LambdaQueryWrapper<WarehouseLineDO>()
                        .eq(WarehouseLineDO::getStartWarehouseId, 1)
                        .eq(WarehouseLineDO::getDestWarehouseId, 2)
                );

                //广州到Kano的线路
                List<WarehouseLineDO> destList = warehouseLineMapper.selectList(new LambdaQueryWrapper<WarehouseLineDO>()
                        .eq(WarehouseLineDO::getStartWarehouseId, 1)
                        .eq(WarehouseLineDO::getDestWarehouseId, 24)
                );
                copyByLine(sourceList, destList);
            }
        }).start();


        new Thread(new Runnable() {
            @Override
            public void run() {
                //义乌到alaba的线路
                List<WarehouseLineDO> sourceList1 = warehouseLineMapper.selectList(new LambdaQueryWrapper<WarehouseLineDO>()
                        .eq(WarehouseLineDO::getStartWarehouseId, 3)
                        .eq(WarehouseLineDO::getDestWarehouseId, 2)
                );

                //义乌到Kano的线路
                List<WarehouseLineDO> destList1 = warehouseLineMapper.selectList(new LambdaQueryWrapper<WarehouseLineDO>()
                        .eq(WarehouseLineDO::getStartWarehouseId, 3)
                        .eq(WarehouseLineDO::getDestWarehouseId, 24)
                );

                copyByLine(sourceList1, destList1);
            }
        }).start();
    }

    @Resource
    private ThreadPoolTaskExecutor threadPoolTaskExecutor;

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void copyV1() {

        Integer[] destArray = {2, 4, 5, 6, 8, 10, 11, 12, 14, 15, 16, 17, 20, 21, 23, 24};
        for (Integer destId : destArray) {
            threadPoolTaskExecutor.execute(new Runnable() {
                @Override
                public void run() {
                    List<WarehouseLineDO> sourceList = warehouseLineMapper.selectList(new LambdaQueryWrapper<WarehouseLineDO>()
                            .eq(WarehouseLineDO::getStartWarehouseId, 1)
                            .eq(WarehouseLineDO::getDestWarehouseId, destId)
                    );

                    //广州到Kano的线路
                    List<WarehouseLineDO> destList = warehouseLineMapper.selectList(new LambdaQueryWrapper<WarehouseLineDO>()
                            .eq(WarehouseLineDO::getStartWarehouseId, 25)
                            .eq(WarehouseLineDO::getDestWarehouseId, destId)
                    );
                    copyByLineV1(sourceList, destList);
                }
            });
        }
    }

    private void copyByLine(List<WarehouseLineDO> sourceList, List<WarehouseLineDO> destList) {
        Map<String, WarehouseLineDO> originMap = sourceList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getTransportType, v -> v));

        for (WarehouseLineDO warehouseLineDO : destList) {
            String transportType = warehouseLineDO.getTransportType();
            if (!originMap.containsKey(transportType)) continue;
            WarehouseLineDO originLine = originMap.get(transportType);

            //查找这个线路的所有路线价格
            List<ProductPriceDO> originPriceList = list(new LambdaQueryWrapper<ProductPriceDO>()
                    .eq(ProductPriceDO::getWarehouseLineId, originLine.getId())
                    .eq(ProductPriceDO::getStatus, CommonStatusEnum.ENABLE.getStatus())
            );

            List<ProductPriceDO> saveList = new ArrayList<>();
            for (ProductPriceDO originPrice : originPriceList) {
                ProductPriceDO copyPrice = new ProductPriceDO();
                BeanUtils.copyProperties(originPrice, copyPrice);
                copyPrice.setId(null);
                copyPrice.setOriginId(originPrice.getId());

                //判断运费和清关费是美元和奈拉，则加
                if (copyPrice.getTransportPriceUnit() == 1L) {
                    BigDecimal yf = copyPrice.getTransportPrice();
                    yf = yf.add(BigDecimal.valueOf(5));
                    copyPrice.setTransportPrice(yf);
                }

                if (copyPrice.getClearancePriceUnit() == 2L) {
                    BigDecimal qgf = copyPrice.getClearancePrice();
                    qgf = qgf.add(BigDecimal.valueOf(30000));
                    copyPrice.setClearancePrice(qgf);
                }
                //修改线路ID
                copyPrice.setWarehouseLineId(warehouseLineDO.getId());

                Date startDate = DateUtil.parse("2023-08-20 00:00:00", "yyyy-MM-dd HH:mm:ss");
                Date endDate = DateUtil.parse("2023-12-31 23:59:59", "yyyy-MM-dd HH:mm:ss");

                copyPrice.setValidateStartDate(startDate);
                copyPrice.setValidateEndDate(endDate);
                copyPrice.setCreateTime(new Date());
                copyPrice.setUpdateTime(new Date());

                saveList.add(copyPrice);
            }

            int batchCount = 1000;
            //sql长度有限制，分批次
            int countBatch = (int) Math.ceil((double) saveList.size() / batchCount);
            List<List<ProductPriceDO>> splitList = Stream.iterate(0, n -> n + 1)
                    .limit(countBatch)
                    .parallel()
                    .map(a -> saveList.stream().skip(a * batchCount).limit(batchCount).parallel().collect(toList())
                    ).collect(toList());
            for (List<ProductPriceDO> savePriceList : splitList) {
                baseMapper.insertBatchSomeColumn(savePriceList);

                //新增特需
                List<Long> originPriceIdList = savePriceList.stream()
                        .map(ProductPriceDO::getOriginId)
                        .collect(toList());
                List<ProductPriceSpecialDO> specialList = productPriceSpecialService.list(new LambdaQueryWrapper<ProductPriceSpecialDO>()
                        .in(ProductPriceSpecialDO::getProductPriceId, originPriceIdList)
                );

                Map<Long, Long> originDestIdMap = savePriceList.stream()
                        .collect(Collectors.toMap(ProductPriceDO::getOriginId, ProductPriceDO::getId));

                if(CollectionUtils.isNotEmpty(specialList)) {
                    List<ProductPriceSpecialDO> copySpecialList = new ArrayList<>();
                    for (ProductPriceSpecialDO originSpecial : specialList) {
                        ProductPriceSpecialDO copySpecial = new ProductPriceSpecialDO();
                        BeanUtils.copyProperties(originSpecial, copySpecial);
                        copySpecial.setId(null);
                        copySpecial.setCreateTime(new Date());
                        copySpecial.setUpdateTime(new Date());
                        copySpecial.setProductPriceId(originDestIdMap.get(originSpecial.getProductPriceId()));

                        copySpecialList.add(copySpecial);
                    }
                    //sql长度有限制，分批次
                    int countBatch1 = (int) Math.ceil((double) copySpecialList.size() / batchCount);
                    List<List<ProductPriceSpecialDO>> splitSpecialList = Stream.iterate(0, n -> n + 1)
                            .limit(countBatch1)
                            .parallel()
                            .map(a -> copySpecialList.stream().skip(a * batchCount).limit(batchCount).parallel().collect(toList())
                            ).collect(toList());
                    for (List<ProductPriceSpecialDO> itemList : splitSpecialList) {
                        productPriceSpecialMapper.insertBatchSomeColumn(itemList);
                    }

                }
            }
        }
    }

    private void copyByLineV1(List<WarehouseLineDO> sourceList, List<WarehouseLineDO> destList) {
        Map<String, WarehouseLineDO> originMap = sourceList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getTransportType, v -> v));

        for (WarehouseLineDO warehouseLineDO : destList) {
            String transportType = warehouseLineDO.getTransportType();
            if(!originMap.containsKey(transportType)) continue;
            WarehouseLineDO originLine = originMap.get(transportType);

            //查找这个线路的所有路线价格
            List<ProductPriceDO> originPriceList = list(new LambdaQueryWrapper<ProductPriceDO>()
                    .eq(ProductPriceDO::getWarehouseLineId, originLine.getId())
                    .eq(ProductPriceDO::getStatus, CommonStatusEnum.ENABLE.getStatus())
            );

            List<ProductPriceDO> saveList = new ArrayList<>();
            for (ProductPriceDO originPrice : originPriceList) {
                ProductPriceDO copyPrice = new ProductPriceDO();
                BeanUtils.copyProperties(originPrice, copyPrice);
                copyPrice.setId(null);
                copyPrice.setOriginId(originPrice.getId());
                //修改线路ID
                copyPrice.setWarehouseLineId(warehouseLineDO.getId());
                copyPrice.setCreateTime(new Date());
                copyPrice.setUpdateTime(new Date());

                saveList.add(copyPrice);
            }

            int batchCount = 1000;
            //sql长度有限制，分批次
            int countBatch = (int) Math.ceil((double) saveList.size() / batchCount);
            List<List<ProductPriceDO>> splitList = Stream.iterate(0, n -> n + 1)
                    .limit(countBatch)
                    .parallel()
                    .map(a -> saveList.stream().skip(a * batchCount).limit(batchCount).parallel().collect(toList())
                    ).collect(toList());
            for (List<ProductPriceDO> savePriceList : splitList) {
                baseMapper.insertBatchSomeColumn(savePriceList);

                //新增特需
                List<Long> originPriceIdList = savePriceList.stream()
                        .map(ProductPriceDO::getOriginId)
                        .collect(toList());
                List<ProductPriceSpecialDO> specialList = productPriceSpecialService.list(new LambdaQueryWrapper<ProductPriceSpecialDO>()
                        .in(ProductPriceSpecialDO::getProductPriceId, originPriceIdList)
                );

                Map<Long, Long> originDestIdMap = savePriceList.stream()
                        .collect(Collectors.toMap(ProductPriceDO::getOriginId, ProductPriceDO::getId));

                if (CollectionUtils.isNotEmpty(specialList)) {
                    List<ProductPriceSpecialDO> copySpecialList = new ArrayList<>();
                    for (ProductPriceSpecialDO originSpecial : specialList) {
                        ProductPriceSpecialDO copySpecial = new ProductPriceSpecialDO();
                        BeanUtils.copyProperties(originSpecial, copySpecial);
                        copySpecial.setId(null);
                        copySpecial.setCreateTime(new Date());
                        copySpecial.setUpdateTime(new Date());
                        copySpecial.setProductPriceId(originDestIdMap.get(originSpecial.getProductPriceId()));

                        copySpecialList.add(copySpecial);
                    }
                    //sql长度有限制，分批次
                    int countBatch1 = (int) Math.ceil((double) copySpecialList.size() / batchCount);
                    List<List<ProductPriceSpecialDO>> splitSpecialList = Stream.iterate(0, n -> n + 1)
                            .limit(countBatch1)
                            .parallel()
                            .map(a -> copySpecialList.stream().skip(a * batchCount).limit(batchCount).parallel().collect(toList())
                            ).collect(toList());
                    for (List<ProductPriceSpecialDO> itemList : splitSpecialList) {
                        productPriceSpecialMapper.insertBatchSomeColumn(itemList);
                    }

                }
            }
        }
    }


    @Override
    public void updateOrderInquiry(ProductPriceUpdateReqVO updateReqVO) {
        //更新单询功能
        Long priceId = updateReqVO.getId();
        ProductPriceDO priceDO = getProductPrice(priceId);
        priceDO.setUpdateTime(null);
        priceDO.setUpdater(null);
        priceDO.setNeedOrderInquiry(updateReqVO.getNeedOrderInquiry());
        productPriceMapper.updateById(priceDO);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void productPriceCopy(String transportType, Long productId, List<Long> productIdList) {
        if (StringUtils.isEmpty(transportType)) return;

        //1.找出所有开通的线路
        List<WarehouseLineDO> list = warehouseService.openedRouterList(new WarehouseLineSearchVO());
        //符合运输方式的线路
        list = list.stream().filter(t -> transportType.equals(t.getTransportType()))
                .collect(toList());


        List<LineChannelVo> channelVoList = new ArrayList<>();
        //2.找出所有渠道列表
        if (transportType.equals("3")) {
            ChannelDataEvent event = new ChannelDataEvent();
            applicationContext.publishEvent(event);

            //空运，价格按照线路和渠道来
            for (WarehouseLineDO lineDO : list) {
                for (Long channelId : event.getChannelIdList()) {
                    channelVoList.add(buildLineChannelVo(lineDO.getId(), channelId));
                }
            }
        } else {
            for (WarehouseLineDO lineDO : list) {
                channelVoList.add(buildLineChannelVo(lineDO.getId(), 0L));
            }
        }
        //3.校验这个商品的所有线路价格是否有正在修改中的
        check(channelVoList, productId);

        List<Long> lineIdList = list.stream()
                .map(WarehouseLineDO::getId)
                .collect(toList());

        //校验通过，执行复制逻辑
        //查找模版商品这个线路的所有路线价格
        List<ProductPriceDO> priceList = list(new LambdaQueryWrapper<ProductPriceDO>()
                .eq(ProductPriceDO::getProductId, productId)
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
        );
        if (CollectionUtil.isEmpty(priceList)) return;

        List<Long> priceIdList = priceList.stream()
                .map(ProductPriceDO::getId)
                .collect(toList());
        ;

        //找到特需和阶梯数据
        List<ProductPriceSpecialDO> priceSpecialList = productPriceSpecialMapper.selectList(new LambdaQueryWrapperX<ProductPriceSpecialDO>()
                .in(ProductPriceSpecialDO::getProductPriceId, priceIdList)
        );

        List<ProductPriceStepDO> priceStepList = productPriceStepMapper.selectList(new LambdaQueryWrapperX<ProductPriceStepDO>()
                .in(ProductPriceStepDO::getProductPriceId, priceIdList)
        );

        //删除商品原有线路价格，拷贝新的线路价格
        for (Long id : productIdList) {
            deleteAndCopy(id, lineIdList, priceList, priceSpecialList, priceStepList);
        }

    }

    /**
     * 删除商品原有的线路价格，并且复制
     */
    private void deleteAndCopy(Long id, List<Long> lineIdList, List<ProductPriceDO> priceList, List<ProductPriceSpecialDO> priceSpecialList, List<ProductPriceStepDO> priceStepList) {
//        baseMapper.delete(new LambdaQueryWrapper<ProductPriceDO>()
//                .eq(ProductPriceDO::getProductId, id)
//                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
//        );
        List<ProductPriceDO> oldPriceList = baseMapper.selectList(new LambdaQueryWrapperX<ProductPriceDO>()
                .eq(ProductPriceDO::getProductId, id)
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
        );

        List<Long> productPriceIdList = oldPriceList.stream()
                .map(ProductPriceDO::getId)
                .collect(toList());
        //删除旧的价格数据
        if (CollectionUtil.isNotEmpty(productPriceIdList)) {
            productPriceMapper.deleteBatchIds(productPriceIdList);
            productPriceSpecialMapper.physicsDelete(Joiner.on(",").join(productPriceIdList));
            productPriceStepMapper.physicsDelete(Joiner.on(",").join(productPriceIdList));
        }

//        Map<Long, List<ProductPriceSpecialDO>> specialListMap = priceSpecialList.stream()
//                .collect(Collectors.groupingBy(ProductPriceSpecialDO::getProductPriceId));
//
//        Map<Long, List<ProductPriceStepDO>> stepListMap = priceStepList.stream()
//                .collect(Collectors.groupingBy(ProductPriceStepDO::getProductPriceId));

        List<ProductPriceDO> insertPriceList = new ArrayList<>();

        for (ProductPriceDO productPriceDO : priceList) {
            productPriceDO.setOldId(productPriceDO.getId());
            productPriceDO.setId(null);
            productPriceDO.setProductId(id);
            productPriceDO.setUpdateTime(new Date());
            insertPriceList.add(productPriceDO);
        }

        baseMapper.insertBatch(insertPriceList);

        Map<Long, ProductPriceDO> priceIdMap = insertPriceList.stream()
                .collect(Collectors.toMap(ProductPriceDO::getOldId, v -> v));

        List<ProductPriceSpecialDO> insertPriceSpecialList = new ArrayList<>();

        for (ProductPriceSpecialDO priceSpecialDO : priceSpecialList) {
            Long priceId = priceSpecialDO.getProductPriceId();
            ProductPriceDO productPriceDO = priceIdMap.get(priceId);
            priceSpecialDO.setId(null);
            priceSpecialDO.setProductPriceId(productPriceDO.getId());
            priceSpecialDO.setUpdateTime(new Date());
            insertPriceSpecialList.add(priceSpecialDO);
        }
        productPriceSpecialMapper.insertBatch(insertPriceSpecialList);

        List<ProductPriceStepDO> insertPriceStepList = new ArrayList<>();
        for (ProductPriceStepDO priceStepDO : priceStepList) {
            Long priceId = priceStepDO.getProductPriceId();
            ProductPriceDO productPriceDO = priceIdMap.get(priceId);
            priceStepDO.setId(null);
            priceStepDO.setProductPriceId(productPriceDO.getId());
            priceStepDO.setUpdateTime(new Date());
            insertPriceStepList.add(priceStepDO);
        }
        productPriceStepMapper.insertBatch(insertPriceStepList);
    }

    private LineChannelVo buildLineChannelVo(Long lineId, Long shippingChannelId) {
        LineChannelVo lineChannelVo = new LineChannelVo();
        lineChannelVo.setLineId(lineId);
        lineChannelVo.setShippingChannelId(shippingChannelId);
        return lineChannelVo;
    }

    @Override
    public AirClearancePriceResultDto queryAirClearancePrice(Long productId, String packaging, Long channelId, BigDecimal weight, BigDecimal volume, BigDecimal quantity) {
        AirClearancePriceResultDto airClearancePriceResultDto = new AirClearancePriceResultDto();

        log.info("空运清关费计算---参数为:{}--{}--{}--{}--{}--{}", productId, packaging, channelId, weight, volume, quantity);
        ProductDO productDO = productService.getProduct(productId);
        if (StringUtils.isBlank(packaging)) {
            packaging = productDO.getPackaging();
        }

        //找出商品的清关费
        List<ProductPriceStepClearanceDO> productClearanceList = productPriceStepClearanceService.list(new LambdaQueryWrapper<ProductPriceStepClearanceDO>()
                .eq(ProductPriceStepClearanceDO::getProductId, productId)
        );

        //1-未设置商品清关费, 2-区间未匹配上
        int productPriceEmptyType = 0;

        BasePrice productPrice = null;
        if (CollectionUtil.isNotEmpty(productClearanceList)) {
            productPrice = matchProductPrice(weight, volume, quantity, productClearanceList);
            if (productPrice == null) {
                productPriceEmptyType = 2;
            }
        } else {
            productPriceEmptyType = 1;
            log.info("空运清关费计算---未设置商品清关费");
        }

        BasePrice channelPrice = null;
        //找出渠道清关费
        int channelPriceEmptyType = 0;
        if (Objects.nonNull(channelId) && channelId > 0L) {
            if (StringUtils.isNotBlank(packaging)) {
                log.info("空运清关费计算---未设置包装类型");
                ChannelPriceEvent event = new ChannelPriceEvent();
                event.setChannelId(channelId);
                event.setPackingType(packaging);
                applicationContext.publishEvent(event);
                List<ChannelClearancePriceDto> channelClearancePriceDtoList = event.getChannelClearancePriceDtoList();
                if (!Objects.isNull(channelClearancePriceDtoList)) {
                    channelPrice = matchChannelPrice(weight, volume, quantity, channelClearancePriceDtoList);
                    if (channelPrice == null) {
                        channelPriceEmptyType = 2;
                    }
                } else {
                    channelPriceEmptyType = 1;
                    log.info("空运清关费计算---未找到包装类型下的渠道清关费");
                }
            } else {
                channelPriceEmptyType = 3;
            }
        }
        BasePrice clearancePrice;
        airClearancePriceResultDto.setProductPriceEmptyType(productPriceEmptyType);
        airClearancePriceResultDto.setChannelPriceEmptyType(channelPriceEmptyType);
        if (productPrice != null && channelPrice != null) {
            int result = productPrice.getClearancePrice().compareTo(channelPrice.getClearancePrice());
            clearancePrice = result >= 0 ? productPrice : channelPrice;
        } else {
            if (productPrice != null) {
                clearancePrice = productPrice;
            } else {
                clearancePrice = channelPrice;
            }
        }
        airClearancePriceResultDto.setClearancePrice(clearancePrice);
        return airClearancePriceResultDto;
    }

    @Override
    public AirClearancePriceResultDto queryAirClearancePrice(Long productId, List<String> packages, Long channelId, BigDecimal weight, BigDecimal volume, BigDecimal quantity) {
        AirClearancePriceResultDto airClearancePriceResultDto = new AirClearancePriceResultDto();

        log.info("空运清关费计算---参数为:{}--{}--{}--{}--{}--{}", productId, packages, channelId, weight, volume, quantity);
        ProductDO productDO = productService.getProduct(productId);
        if (CollectionUtil.isEmpty(packages) && StringUtils.isNotBlank(productDO.getPackaging())) {
            packages = Collections.singletonList(productDO.getPackaging());
        }

        //找出商品的清关费
        List<ProductPriceStepClearanceDO> productClearanceList = productPriceStepClearanceService.list(new LambdaQueryWrapper<ProductPriceStepClearanceDO>()
                .eq(ProductPriceStepClearanceDO::getProductId, productId)
        );

        //1-未设置商品清关费, 2-区间未匹配上
        int productPriceEmptyType = 0;
        boolean isAirWeightLimit = Objects.nonNull(productDO.getAirWeightLimit()) && productDO.getAirWeightLimit().compareTo(weight) < 0;

        BasePrice productPrice = null;
        if (CollectionUtil.isNotEmpty(productClearanceList)) {
            productPrice = matchProductPrice(weight, volume, quantity, productClearanceList);
            if (productPrice == null) {
                productPriceEmptyType = 2;
            }
        } else {
            productPriceEmptyType = 1;
            log.info("空运清关费计算---未设置商品清关费");
        }

        BasePrice channelPrice = null;
        //找出渠道清关费
        int channelPriceEmptyType = 0;
        if (Objects.nonNull(channelId) && channelId > 0L) {
            if (productPriceEmptyType == 0) {
                // 只有当商品设置了空运清关费时，才需要去查询渠道的清关费，然后对两者进行比价，否则就时清关费未报价异常
                if (CollectionUtil.isNotEmpty(packages)) {
                    log.info("空运清关费计算---未设置包装类型");
                    ChannelPriceEvent event = new ChannelPriceEvent();
                    event.setChannelId(channelId);
                    event.setPackingTypes(packages);
                    applicationContext.publishEvent(event);
                    List<ChannelClearancePriceDto> channelClearancePriceDtoList = event.getChannelClearancePriceDtoList();
                    if (!Objects.isNull(channelClearancePriceDtoList)) {
                        channelPrice = matchChannelPrice(weight, volume, quantity, channelClearancePriceDtoList);
                        if (channelPrice == null) {
                            channelPriceEmptyType = 2;
                        }
                    } else {
                        channelPriceEmptyType = 1;
                        log.info("空运清关费计算---未找到包装类型下的渠道清关费");
                    }
                } else {
                    channelPriceEmptyType = 3;
                }
            }
        }
        BasePrice clearancePrice;
        airClearancePriceResultDto.setProductPriceEmptyType(productPriceEmptyType);
        airClearancePriceResultDto.setChannelPriceEmptyType(channelPriceEmptyType);
        if (productPrice != null && channelPrice != null) {
            int result = 0;
            if (productPrice.getClearancePriceUnit() != channelPrice.getClearancePriceUnit()) {
                ExchangeRateRespDTO productExchangeRate = currencyApi.getCurrencyRate(productPrice.getClearancePriceUnit(), 1L);
                BigDecimal productClearancePrice = productExchangeRate.getCurrencyRate().multiply(productPrice.getClearancePrice()).setScale(2, BigDecimal.ROUND_HALF_UP);
                ExchangeRateRespDTO channelExchangeRate = currencyApi.getCurrencyRate(channelPrice.getClearancePriceUnit(), 1L);
                BigDecimal channelClearancePrice = channelExchangeRate.getCurrencyRate().multiply(channelPrice.getClearancePrice()).setScale(2, BigDecimal.ROUND_HALF_UP);
                result = productClearancePrice.compareTo(channelClearancePrice);
            }else {
                result = productPrice.getClearancePrice().compareTo(channelPrice.getClearancePrice());
            }
            clearancePrice = result >= 0 ? productPrice : channelPrice;
        } else {
            if (productPrice != null) {
                clearancePrice = productPrice;
            } else {
                clearancePrice = channelPrice;
            }
        }
        airClearancePriceResultDto.setIsAirWeightLimit(isAirWeightLimit);
        airClearancePriceResultDto.setClearancePrice(clearancePrice);
        return airClearancePriceResultDto;
    }

    private BasePrice matchProductPrice(BigDecimal weight, BigDecimal volume, BigDecimal quantity, List<ProductPriceStepClearanceDO> productClearanceList) {
//        // 获取最小的
//        ProductPriceStepClearanceDO stepClearanceDO = productClearanceList.stream().min(Comparator.comparing(ProductPriceStepClearanceDO::getStartNum)).orElse(null);
//        if (Objects.nonNull(stepClearanceDO)){
//            // 当计量值为0时，以最小计量阶梯价做单价
//            BigDecimal unitValue = getUnitValue(weight, volume, quantity, stepClearanceDO.getWeightUnit());
//            if (unitValue.compareTo(stepClearanceDO.getStartNum()) <= 0){
//                BasePrice basePrice = new BasePrice();
//                BeanUtils.copyProperties(stepClearanceDO, basePrice);
//                basePrice.setClearanceVolumeUnit(stepClearanceDO.getWeightUnit());
//                basePrice.setAirClearanceSource(1);
//                basePrice.setAirClearanceInfo(JSONObject.toJSONString(stepClearanceDO));
//                return basePrice;
//            }
//        }
        BasePrice basePrice = new BasePrice();
        for (ProductPriceStepClearanceDO productPriceStepClearanceDO : productClearanceList) {
            basePrice.setClearancePrice(BigDecimal.ZERO);
            basePrice.setClearancePriceUnit(productPriceStepClearanceDO.getClearancePriceUnit());
            basePrice.setClearanceVolumeUnit(productPriceStepClearanceDO.getWeightUnit());
            if (matchPrice(productPriceStepClearanceDO.getWeightUnit(), productPriceStepClearanceDO.getStartNum(), productPriceStepClearanceDO.getEndNum(), weight, volume, quantity)) {
                BeanUtils.copyProperties(productPriceStepClearanceDO, basePrice);
                basePrice.setClearanceVolumeUnit(productPriceStepClearanceDO.getWeightUnit());
                basePrice.setAirClearanceSource(1);
                basePrice.setAirClearanceInfo(JSONObject.toJSONString(productPriceStepClearanceDO));
                return basePrice;
            }
        }
        log.info("空运清关费计算---商品清关费用区间未匹配上");
        return basePrice;
    }

    private BasePrice matchChannelPrice(BigDecimal weight, BigDecimal volume, BigDecimal quantity, List<ChannelClearancePriceDto> channelClearancePriceDtoList) {
//        // 获取最小的
//        ChannelClearancePriceDto clearancePriceDto = channelClearancePriceDtoList.stream().min(Comparator.comparing(ChannelClearancePriceDto::getStartNum)).orElse(null);
//        if (Objects.nonNull(clearancePriceDto)){
//            // 当计量值为0时，以最小计量阶梯价做单价
//            BigDecimal unitValue = getUnitValue(weight, volume, quantity, clearancePriceDto.getWeightUnit());
//            if (unitValue.compareTo(clearancePriceDto.getStartNum()) <= 0){
//                BasePrice basePrice = new BasePrice();
//                BeanUtils.copyProperties(clearancePriceDto, basePrice);
//                basePrice.setClearanceVolumeUnit(clearancePriceDto.getWeightUnit());
//                basePrice.setAirClearanceSource(1);
//                basePrice.setAirClearanceInfo(JSONObject.toJSONString(clearancePriceDto));
//                return basePrice;
//            }
//        }
        BasePrice basePrice = new BasePrice();
        for (ChannelClearancePriceDto channelClearancePriceDto : channelClearancePriceDtoList) {
            basePrice.setClearancePrice(BigDecimal.ZERO);
            basePrice.setClearancePriceUnit(channelClearancePriceDto.getClearancePriceUnit());
            basePrice.setClearanceVolumeUnit(channelClearancePriceDto.getWeightUnit());
            //首先校验重量单位是否匹配
            if (matchPrice(channelClearancePriceDto.getWeightUnit(), channelClearancePriceDto.getStartNum(), channelClearancePriceDto.getEndNum(), weight, volume, quantity)) {
                BeanUtils.copyProperties(channelClearancePriceDto, basePrice);
                basePrice.setClearanceVolumeUnit(channelClearancePriceDto.getWeightUnit());
                basePrice.setAirClearanceSource(2);
                basePrice.setAirClearanceInfo(JSONObject.toJSONString(channelClearancePriceDto));
                return basePrice;
            }
        }
        log.info("空运清关费计算---渠道清关费用区间未匹配上");
        return basePrice;
    }

    private boolean matchPrice(Long unit, BigDecimal startWeight, BigDecimal endWeight, BigDecimal weight, BigDecimal volume, BigDecimal quantity) {
        //首先校验重量单位是否匹配
        BigDecimal unitValue = getUnitValue(weight, volume, quantity, unit);
        if (unitValue == null) return false;

        if (startWeight == null || endWeight == null) {
            return false;
        }
        return unitValue.compareTo(startWeight) >= 0 && unitValue.compareTo(endWeight) <= 0;
    }

    private BigDecimal getUnitValue(BigDecimal weight, BigDecimal volume, BigDecimal quantity, Long unit) {
        //6--KG   7-m3   10-个
        if (unit == 6L) return weight;
        if (unit == 7L) return volume;
        if (unit == 10L) return quantity;
        return null;
    }

    @Override
    public void batchAddPrice(ProductPriceBatchAddPriceReqVo batchAddPriceReqVo, ProductPriceCheckAO productPriceCheckAO) {
        String json = JSON.toJSONString(batchAddPriceReqVo);
        List<ProductDO> productList = productPriceCheckAO.getProductList();
        List<LineChannelVo> lineChannelList = batchAddPriceReqVo.getLineChannelList();
        logAndDelKey(json, productList, lineChannelList);

        List<WarehouseLineDO> warehouseLineList = productPriceCheckAO.getWarehouseLineList();
        String content = productPriceCheckAO.getContent();

        int countBatch = (int) Math.ceil((double) productList.size() / DATA_THRESHOLD);
        List<List<ProductDO>> splitList = Stream.iterate(0, n -> n + 1)
                .limit(countBatch)
                .parallel()
                .map(a -> productList.stream().skip(a * DATA_THRESHOLD).limit(DATA_THRESHOLD).parallel().collect(toList())
                ).collect(toList());
        for (List<ProductDO> itemList : splitList) {
            ProductPriceCreateReqVO createReqVO = new ProductPriceCreateReqVO();
            BeanUtils.copyProperties(batchAddPriceReqVo, createReqVO);
            batchAddPrice(createReqVO, warehouseLineList, itemList);
        }
        //发送站内信
        sendMessage(content);
    }

    @Transactional(rollbackFor = Exception.class)
    public void batchAddPrice(ProductPriceCreateReqVO batchAddPriceReqVo, List<WarehouseLineDO> warehouseLineList, List<ProductDO> productList) {
        long t1 = System.currentTimeMillis();
        Long loginUserId = SecurityFrameworkUtils.getLoginUserId();

        ProductPriceDO productPrice = ProductPriceConvert.INSTANCE.convert(batchAddPriceReqVo);

        List<LineChannelVo> lineChannelList = batchAddPriceReqVo.getLineChannelList();
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());
        Set<Long> channelIdList = lineChannelList.stream()
                .filter(t -> t.getShippingChannelId() != null)
                .map(LineChannelVo::getShippingChannelId)
                .collect(Collectors.toSet());

        Map<Long, String> lineTransportTypeMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, WarehouseLineDO::getTransportType));

        List<Long> idList = productList.stream()
                .map(ProductDO::getId)
                .collect(toList());
        //这个产品这么多线路已有的价格
        List<ProductPriceDO> dbPriceList = productPriceMapper.selectList(new LambdaQueryWrapperX<ProductPriceDO>()
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
                .in(CollectionUtil.isNotEmpty(channelIdList), ProductPriceDO::getShippingChannelId, channelIdList)
                .in(ProductPriceDO::getProductId, idList));
        Map<String, ProductPriceDO> productLineChannelMap = dbPriceList.stream()
                .collect(Collectors.toMap(t -> t.getProductId() + "-" + t.getWarehouseLineId() + "-" + t.getShippingChannelId(), v -> v, (v1, v2) -> v2));

        List<ProductPriceDO> priceList = new ArrayList<>();
        List<ProductPriceDO> updateList = new ArrayList<>();

        List<ProductPriceDO> snapshotList = new ArrayList<>();

        Date current = new Date();
        for (Long productId : idList) {
            for (LineChannelVo lineChannel : lineChannelList) {
                Long lineId = lineChannel.getLineId();
                Long shippingChannelId = lineChannel.getShippingChannelId();
                if(shippingChannelId == null) shippingChannelId = 0L;
                if (shippingChannelId > 0) {
                    batchAddPriceReqVo.setStepPrice(1);
                }

                String key = productId + "-" + lineId + "-" + shippingChannelId;

                ProductPriceDO dbItem;
                if (productLineChannelMap.containsKey(key)) {
                    dbItem = productLineChannelMap.get(key);
                    ProductPriceDO snapshot = ObjectUtil.clone(dbItem);
                    priceList.add(dbItem);

                    dbItem.setUpdateTime(current);
                    boolean update = addPrice(productPrice.getPriceType(), dbItem, productPrice);

                    if (update) {
                        dbItem.setUpdater(String.valueOf(loginUserId));
                        updateList.add(dbItem);
                        //添加日志
                        snapshotList.add(snapshot);
                    }
                }
            }
        }
        if (CollectionUtil.isEmpty(priceList)) {
            long t2 = System.currentTimeMillis();
            log.info("批量加价路线处理耗时:" + (t2 - t1));
            log.warn("没有价格需要修改");
            deleteKey(idList, lineChannelList);
            return;
        }

        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        priceList.forEach(t -> priceTransportTypeMap.put(t.getId(), lineTransportTypeMap.get(t.getWarehouseLineId())));

        if (null != batchAddPriceReqVo.getStepPrice() && batchAddPriceReqVo.getStepPrice() == 1) {
            snapshotList = new ArrayList<>();
            updateList = new ArrayList<>();

            ProductPriceStepObjectAO freightAO = stepListAddPriceAir(priceList, batchAddPriceReqVo.getFreightPriceStepList(), PriceTypeEnum.FREIGHT, priceTransportTypeMap, snapshotList, updateList);
            ProductPriceStepObjectAO clearanceAO = stepListAddPriceAir(priceList, batchAddPriceReqVo.getClearancePriceStepList(), PriceTypeEnum.CLEARANCE, priceTransportTypeMap, snapshotList, updateList);
            ProductPriceStepObjectAO fullAO = stepListAddPriceAir(priceList, batchAddPriceReqVo.getFullPriceStepList(), PriceTypeEnum.FULL_PRICE, priceTransportTypeMap, snapshotList, updateList);

            //添加修改日志,需要查询特需和阶梯，所以需要在修改之前执行
            batchAddSnapshot(snapshotList);

            if(CollectionUtil.isNotEmpty(updateList)) {
                productPriceMapper.updateBatch(updateList);
            }

            List<ProductPriceStepDO> stepUpdateList = new ArrayList<>();
            List<ProductPriceStepSpecialDO> stepSpecialUpdateList = new ArrayList<>();
            List<ProductPriceStepPackagingDO> stepPackagingUpdateList = new ArrayList<>();
            addStepList(freightAO, stepUpdateList, stepSpecialUpdateList, stepPackagingUpdateList);
            addStepList(clearanceAO, stepUpdateList, stepSpecialUpdateList, stepPackagingUpdateList);
            addStepList(fullAO, stepUpdateList, stepSpecialUpdateList, stepPackagingUpdateList);

            if (CollectionUtil.isNotEmpty(stepUpdateList)) {
                productPriceStepMapper.updateBatch(stepUpdateList);
            }

            if(CollectionUtil.isNotEmpty(stepSpecialUpdateList)) {
                productPriceStepSpecialMapper.updateBatch(stepSpecialUpdateList);
            }

            if(CollectionUtil.isNotEmpty(stepPackagingUpdateList)) {
                productPriceStepPackagingMapper.updateBatch(stepPackagingUpdateList);
            }

        } else {
            if (CollectionUtil.isNotEmpty(updateList)) {
                productPriceMapper.updateBatch(updateList);

                //添加修改日志,需要查询特需和阶梯，所以需要在修改之前执行
                batchAddSnapshot(snapshotList);

                //价格特需
                specialListAddPrice(updateList, batchAddPriceReqVo.getSpecialList(), priceTransportTypeMap);

                //价格包装单位
                List<ProductPricePackagingDO> allPackagingList = buildAllPackageList(batchAddPriceReqVo);
                packagingListAddPrice(updateList, allPackagingList, priceTransportTypeMap);
            }
        }

        long t2 = System.currentTimeMillis();
        log.info("批量加价路线处理耗时:" + (t2 - t1));
        deleteKey(idList, lineChannelList);
    }

    private boolean addPrice(Integer priceType, BasePrice dbItem, BasePrice addPrice) {
        boolean update = false;
        if (priceType == PriceTypeEnum.FREIGHT.getPriceType()) {
            // 清关费+运费 清空全包价设置
            BigDecimal yf = dbItem.getTransportPrice();
            Long yfPriceUnit = dbItem.getTransportPriceUnit();
            Long yfVolumeUnit = dbItem.getTransportVolumeUnit();

            BigDecimal addYf = addPrice.getTransportPrice();
            Long addYfPriceUnit = addPrice.getTransportPriceUnit();
            Long addYfVolumeUnit = addPrice.getTransportVolumeUnit();

            if (addYf != null && addYfPriceUnit != null && addYfVolumeUnit != null) {
                if (yfPriceUnit.equals(addYfPriceUnit) && yfVolumeUnit.equals(addYfVolumeUnit)) {
                    update = true;
                    yf = yf.add(addYf);
                    dbItem.setTransportPrice(yf);
                }
            }

            BigDecimal qgf = dbItem.getClearancePrice();
            Long qgfPriceUnit = dbItem.getClearancePriceUnit();
            Long qgfVolumeUnit = dbItem.getClearanceVolumeUnit();

            BigDecimal addQgf = addPrice.getClearancePrice();
            Long addQgfPriceUnit = addPrice.getClearancePriceUnit();
            Long addQgfVolumeUnit = addPrice.getClearanceVolumeUnit();
            if (addQgf != null && addQgfPriceUnit != null && addQgfVolumeUnit != null) {
                if (qgfPriceUnit.equals(addQgfPriceUnit) && qgfVolumeUnit.equals(addQgfVolumeUnit)) {
                    update = true;
                    qgf = qgf.add(addQgf);
                    dbItem.setClearancePrice(qgf);
                }
            }
        }
        if (priceType == PriceTypeEnum.FULL_PRICE.getPriceType()) {
            BigDecimal qbj = dbItem.getAllPrice();
            Long qbjPriceUnit = dbItem.getAllPriceUnit();
            Long qbjVolumeUnit = dbItem.getAllVolumeUnit();

            BigDecimal addQbj = addPrice.getAllPrice();
            Long addQbjPriceUnit = addPrice.getAllPriceUnit();
            Long addQbjVolumeUnit = addPrice.getAllVolumeUnit();
            if (addQbj != null && addQbjPriceUnit != null && addQbjVolumeUnit != null) {
                if (qbjPriceUnit.equals(addQbjPriceUnit) && qbjVolumeUnit.equals(addQbjVolumeUnit)) {
                    update = true;
                    qbj = qbj.add(addQbj);
                    dbItem.setAllPrice(qbj);
                }
            }
        }
        if(priceType == PriceTypeEnum.CLEARANCE.getPriceType()) {
            BigDecimal qgf = dbItem.getClearancePrice();
            Long qgfPriceUnit = dbItem.getClearancePriceUnit();
            Long qgfVolumeUnit = dbItem.getClearanceVolumeUnit();

            BigDecimal addQgf = addPrice.getClearancePrice();
            Long addQgfPriceUnit = addPrice.getClearancePriceUnit();
            Long addQgfVolumeUnit = addPrice.getClearanceVolumeUnit();
            if (addQgf != null && addQgfPriceUnit != null && addQgfVolumeUnit != null) {
                if (qgfPriceUnit.equals(addQgfPriceUnit) && qgfVolumeUnit.equals(addQgfVolumeUnit)) {
                    update = true;
                    qgf = qgf.add(addQgf);
                    dbItem.setClearancePrice(qgf);
                }
            }
        }
        return update;
    }

    private void specialListAddPrice(List<ProductPriceDO> productPriceList, List<ProductPriceSpecialDO> specialList, Map<Long, String> priceTransportTypeMap) {
        if (specialList == null) return;

        specialList = specialList.stream()
                .filter(t -> t.getTransportPrice() != null ||
                        t.getClearancePrice() != null ||
                        t.getAllPrice() != null)
                .collect(Collectors.toList());

        if (CollectionUtil.isEmpty(specialList)) return;

        List<ProductPriceSpecialDO> allSpecialList = new ArrayList<>();
        List<List<ProductPriceDO>> idPartList = Lists.partition(productPriceList, 1000);
        for (List<ProductPriceDO> priceItemList : idPartList) {
            List<Long> idList = priceItemList.stream()
                    .map(ProductPriceDO::getId)
                    .collect(toList());
            List<ProductPriceSpecialDO> itemSpecialList = productPriceSpecialService.selectList(new LambdaQueryWrapper<ProductPriceSpecialDO>()
                    .in(ProductPriceSpecialDO::getProductPriceId, idList)
            );
            allSpecialList.addAll(itemSpecialList);
        }
        Map<Long, List<ProductPriceSpecialDO>> priceSpecialMap = allSpecialList.stream()
                .collect(Collectors.groupingBy(ProductPriceSpecialDO::getProductPriceId));

        List<ProductPriceSpecialDO> updateList = new ArrayList<>();

        Map<String, ProductPriceSpecialDO> specialMap = specialList.stream()
                .collect(Collectors.toMap(ProductPriceSpecialDO::getSpecialDictType, v -> v));

        for (ProductPriceDO productPriceDO : productPriceList) {
            Long productPriceId = productPriceDO.getId();
            List<ProductPriceSpecialDO> productSpecialList = priceSpecialMap.get(productPriceId);
            //价格没有特需则跳过
            if (CollectionUtil.isEmpty(productSpecialList)) continue;

            for (ProductPriceSpecialDO dbSpecial : productSpecialList) {
                String specialType = dbSpecial.getSpecialDictType();
                ProductPriceSpecialDO updateSpecial = specialMap.get(specialType);
                if (updateSpecial == null) continue;
                String transportType = priceTransportTypeMap.get(productPriceId);
                initBasePrice(updateSpecial, transportType);
                boolean update = addPrice(productPriceDO.getPriceType(), dbSpecial, updateSpecial);
                if (update) {
                    updateList.add(dbSpecial);
                }
            }
        }
        //首先删除所有的特殊需求列表
        if (CollectionUtil.isNotEmpty(updateList)) {
            productPriceSpecialMapper.updateBatch(updateList);
        }
    }

    private void packagingListAddPrice(List<ProductPriceDO> productPriceList, List<ProductPricePackagingDO> packagingList, Map<Long, String> priceTransportTypeMap) {
        if (packagingList == null) return;

        packagingList = packagingList.stream()
                .filter(t -> t.getPackagingPrice() != null)
                .collect(Collectors.toList());

        if (CollectionUtil.isEmpty(packagingList)) return;

        List<ProductPricePackagingDO> allPackagingList = new ArrayList<>();
        List<List<ProductPriceDO>> idPartList = Lists.partition(productPriceList, 1000);
        for (List<ProductPriceDO> priceItemList : idPartList) {
            List<Long> idList = priceItemList.stream()
                    .map(ProductPriceDO::getId)
                    .collect(toList());
            List<ProductPricePackagingDO> itemPackagingList = productPricePackagingMapper.selectList(new LambdaQueryWrapper<ProductPricePackagingDO>()
                    .in(ProductPricePackagingDO::getProductPriceId, idList)
            );
            allPackagingList.addAll(itemPackagingList);
        }
        Map<Long, List<ProductPricePackagingDO>> pricePackagingMap = allPackagingList.stream()
                .collect(Collectors.groupingBy(ProductPricePackagingDO::getProductPriceId));

        List<ProductPricePackagingDO> updateList = new ArrayList<>();

        for (ProductPriceDO productPriceDO : productPriceList) {
            Long productPriceId = productPriceDO.getId();
            List<ProductPricePackagingDO> productPackagingList = pricePackagingMap.get(productPriceId);
            //价格没有包装单位则跳过
            if (CollectionUtil.isEmpty(productPackagingList)) continue;

            for (ProductPricePackagingDO dbPackaging: productPackagingList) {
                for (ProductPricePackagingDO addPricePackaging : packagingList) {
                    Integer priceType = addPricePackaging.getPriceType();
                    if(priceType == null) continue;
                    boolean update = checkPackageAddPrice(addPricePackaging, dbPackaging) && priceType.equals(dbPackaging.getPriceType());
                    if(update) {
                        BigDecimal price = dbPackaging.getPackagingPrice();
                        price = price.add(addPricePackaging.getPackagingPrice());
                        dbPackaging.setPackagingPrice(price);
                        updateList.add(dbPackaging);
                    }
                }
            }
        }
        //首先删除所有的特殊需求列表
        if (CollectionUtil.isNotEmpty(updateList)) {
            productPricePackagingMapper.updateBatch(updateList);
        }
    }

    private boolean checkPackageAddPrice(BasePackagingPrice addPricePackagingDO, BasePackagingPrice dbStepPackaging) {
        Long addPricePackagingPriceUnit = addPricePackagingDO.getPackagingPriceUnit();
        Long addPricePackagingVolumeUnit = addPricePackagingDO.getPackagingVolumeUnit();
        return checkPackageEquals(addPricePackagingDO, dbStepPackaging)
                && dbStepPackaging.getPackagingPriceUnit().equals(addPricePackagingPriceUnit)
                && dbStepPackaging.getPackagingVolumeUnit().equals(addPricePackagingVolumeUnit);
    }

    private boolean checkPackageEquals(BasePackagingPrice addPricePackagingDO, BasePackagingPrice dbStepPackaging) {
        String addPricePackagingTypes = addPricePackagingDO.getPackagingTypes();
        if(StringUtils.isBlank(addPricePackagingTypes)) return false;
        Long addPricePackagingPriceUnit = addPricePackagingDO.getPackagingPriceUnit();
        if(addPricePackagingPriceUnit == null) return false;
        Long addPricePackagingVolumeUnit = addPricePackagingDO.getPackagingVolumeUnit();
        if(addPricePackagingVolumeUnit == null) return false;
        Set<String> addPriceTypeSets = new HashSet<>(Arrays.asList(addPricePackagingTypes.split(",")));
        Set<String> dbPriceTypeSets = new HashSet<>(Arrays.asList(dbStepPackaging.getPackagingTypes().split(",")));

        return addPriceTypeSets.equals(dbPriceTypeSets);
    }

    @Override
    public Long countByProductIdAndTransportType(Long productId, List<String> transportTypeList) {
        return productPriceMapper.countByProductIdAndTransportType(productId, transportTypeList);
    }

    @Override
    public Long channelPriceCount(Long channelId) {
        return count(new LambdaQueryWrapper<ProductPriceDO>()
                .eq(ProductPriceDO::getShippingChannelId, channelId)
        );
    }

    @Transactional(rollbackFor = Exception.class)
    public void batchUpdateAir(ProductPriceAirCreateReqVO createReqVO, List<WarehouseLineDO> warehouseLineList, List<ProductDO> productList) {
        long t1 = System.currentTimeMillis();
        Long loginUserId = SecurityFrameworkUtils.getLoginUserId();


        ProductPriceDO productPrice = ProductPriceConvert.INSTANCE.convert(createReqVO);
        productPrice.setStepPrice(1);

        checkAndUpdateStepToMainPrice(productPrice, createReqVO);

        List<LineChannelVo> lineChannelList = createReqVO.getLineChannelList();
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());

        Map<Long, String> lineTransportTypeMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, WarehouseLineDO::getTransportType));

        List<Long> idList = productList.stream()
                .map(ProductDO::getId)
                .collect(toList());
        //这个产品这么多线路已有的价格
        List<ProductPriceDO> dbPriceList = productPriceMapper.selectList(new LambdaQueryWrapperX<ProductPriceDO>()
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
                .in(ProductPriceDO::getProductId, idList));
        Map<String, ProductPriceDO> productLineChannelMap = dbPriceList.stream()
                .collect(Collectors.toMap(t -> t.getProductId() + "-" + t.getWarehouseLineId() + "-" + t.getShippingChannelId(), v -> v, (v1, v2) -> v2));


        List<ProductPriceDO> insertList = new ArrayList<>();
        List<ProductPriceDO> updateList = new ArrayList<>();

        List<ProductPriceDO> snapshotList = new ArrayList<>();

        Date current = new Date();
        for (Long productId : idList) {
            for (LineChannelVo lineChannel : lineChannelList) {
                Long lineId = lineChannel.getLineId();
                Long shippingChannelId = lineChannel.getShippingChannelId();
                String key = productId + "-" + lineId + "-" + shippingChannelId;

                ProductPriceDO dbItem;
                if (productLineChannelMap.containsKey(key)) {
                    dbItem = productLineChannelMap.get(key);

                    //添加日志
                    ProductPriceDO snapshot = ObjectUtil.clone(dbItem);
                    snapshotList.add(snapshot);
                    dbItem.setUpdater(String.valueOf(loginUserId));
                    updateColumn(dbItem, productPrice);
                    updateList.add(dbItem);
                } else {
                    dbItem = ObjectUtil.clone(productPrice);
                    dbItem.setProductId(productId);
                    dbItem.setWarehouseLineId(lineId);
                    dbItem.setShippingChannelId(shippingChannelId);

                    if (dbItem.getNeedBook() == null) {
                        dbItem.setNeedBook(0);
                    }
                    if (dbItem.getPriceType() == null) {
                        dbItem.setPriceType(0);
                    }
                    if (dbItem.getNeedPay() == null) {
                        dbItem.setNeedPay(0);
                    }
                    if (dbItem.getAuditStatus() == null) {
                        dbItem.setAuditStatus(0);
                    }
                    if (dbItem.getPayDelivery() == null) {
                        dbItem.setPayDelivery(0);
                    }
                    if (dbItem.getStepPrice() == null) {
                        dbItem.setStepPrice(0);
                    }
                    if (dbItem.getMinWeight() == null) {
                        dbItem.setMinWeight(BigDecimal.ZERO);
                    }
                    if (dbItem.getMinWeightUnit() == null) {
                        dbItem.setMinWeightUnit(7L);
                    }
                    dbItem.setBlacklist(0);
                    dbItem.setStatus(0);
                    dbItem.setCreateTime(current);
                    insertList.add(dbItem);
                }
                dbItem.setUpdateTime(current);
                String transportType = lineTransportTypeMap.get(lineId);
                initBasePrice(dbItem, transportType);
            }
        }
        if (CollectionUtil.isNotEmpty(insertList)) {
            productPriceMapper.insertBatchSomeColumn(insertList);
        }
        if (CollectionUtil.isNotEmpty(updateList)) {
            productPriceMapper.updateBatch(updateList);
        }
        List<ProductPriceDO> productPriceList = new ArrayList<>();
        productPriceList.addAll(insertList);
        productPriceList.addAll(updateList);
        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        productPriceList.forEach(t -> priceTransportTypeMap.put(t.getId(), lineTransportTypeMap.get(t.getWarehouseLineId())));

        //添加修改日志,需要查询特需和阶梯，所以需要在修改之前执行
        batchAddSnapshot(snapshotList);

        List<Long> productPriceIdList = productPriceList.stream().map(ProductPriceDO::getId).collect(Collectors.toList());

        List<ProductPriceStepDO> allStepList = buildAllStepList(createReqVO);
        clearAndSavePriceStepList(productPriceIdList, allStepList, priceTransportTypeMap, createReqVO.isBatch());

        long t2 = System.currentTimeMillis();
        log.info("路线处理耗时:" + (t2 - t1));
        deleteKey(idList, lineChannelList);

        handleOrderException(productPriceIdList, createReqVO.isSync());
    }

    private void checkStepPrice(List<ProductPriceStepDO> priceStepList, PriceTypeEnum priceTypeEnum) {
        if (CollectionUtil.isNotEmpty(priceStepList)) {
//            priceStepList = priceStepList.stream()
//                    .filter(t -> t.getTransportPrice() != null ||
//                            t.getClearancePrice() != null ||
//                            t.getAllPrice() != null)
//                    .collect(Collectors.toList());

            for (ProductPriceStepDO item : priceStepList) {
                if (item.getRankNum() == null) {
                    throw new ServiceException(500, "阶梯排序不能为空");
                }
                item.setPriceType(priceTypeEnum.getPriceType());
            }
//            switch (priceTypeEnum) {
//                case FREIGHT:
//                    for (ProductPriceStepDO item : priceStepList) {
//                        if (item.getTransportPrice() == null) {
//                            throw new ServiceException(500, "价格不能为空");
//                        }
//                    }
//                    break;
//
//                case CLEARANCE:
//                    for (ProductPriceStepDO item : priceStepList) {
//                        if (item.getClearancePrice() == null) {
//                            throw new ServiceException(500, "价格不能为空");
//                        }
//                    }
//                    break;
//
//                case FULL_PRICE:
//                    for (ProductPriceStepDO item : priceStepList) {
//                        if (item.getAllPrice() == null) {
//                            throw new ServiceException(500, "价格不能为空");
//                        }
//                    }
//                    break;
//            }
        }
    }

    private void checkAndUpdateStepToMainPrice(ProductPriceDO productPrice, ProductPriceReqBaseVO productPriceStepBaseVO) {
        checkStepPrice(productPriceStepBaseVO);
        updateStepToMainPriceAir(productPrice, productPriceStepBaseVO.getFreightPriceStepList(), PriceTypeEnum.FREIGHT);
        updateStepToMainPriceAir(productPrice, productPriceStepBaseVO.getClearancePriceStepList(), PriceTypeEnum.CLEARANCE);
        updateStepToMainPriceAir(productPrice, productPriceStepBaseVO.getFullPriceStepList(), PriceTypeEnum.FULL_PRICE);
    }

    private void updateStepToMainPriceAir(ProductPriceDO productPrice, List<ProductPriceStepDO> priceStepList, PriceTypeEnum priceTypeEnum) {
        if (CollectionUtil.isNotEmpty(priceStepList)) {
            ProductPriceStepDO stepDO = priceStepList.get(0);
            switch (priceTypeEnum) {
                case FREIGHT:
                    if(stepDO.getTransportPrice() != null) {
                        productPrice.setTransportPrice(stepDO.getTransportPrice());
                        productPrice.setTransportPriceUnit(stepDO.getTransportPriceUnit());
                        productPrice.setTransportVolumeUnit(stepDO.getTransportVolumeUnit());
                    }
                    break;

                case CLEARANCE:
                    if(stepDO.getClearancePrice() != null) {
                        productPrice.setClearancePrice(stepDO.getClearancePrice());
                        productPrice.setClearancePriceUnit(stepDO.getClearancePriceUnit());
                        productPrice.setClearanceVolumeUnit(stepDO.getClearanceVolumeUnit());
                    }
                    break;

                case FULL_PRICE:
                    if(stepDO.getAllPrice() != null) {
                        productPrice.setAllPrice(stepDO.getAllPrice());
                        productPrice.setAllPriceUnit(stepDO.getAllPriceUnit());
                        productPrice.setAllVolumeUnit(stepDO.getAllVolumeUnit());
                    }
                    break;
            }
        }
    }


    @Override
    @Transactional(rollbackFor = Exception.class)
    public void updateProductPriceAir(ProductPriceAirUpdateReqVO updateReqVO) {
        // 校验存在
        Long priceId = updateReqVO.getId();
        this.validateProductPriceExists(priceId);
        // 更新
        ProductPriceDO updateObj = ProductPriceConvert.INSTANCE.convert(updateReqVO);
        updateObj.setStepPrice(1);
        BigDecimal lineWeightLimitNew = updateObj.getWeightLimit();

        checkAndUpdateStepToMainPrice(updateObj, updateReqVO);

        ProductPriceDO priceDO = getProductPrice(priceId);
        BigDecimal lineWeightLimitOld = priceDO.getWeightLimit();
        //添加修改日志
        batchAddSnapshot(Collections.singletonList(priceDO));

        productPriceMapper.updateById(updateObj);
        if (updateObj.getValidateStartDate() == null || updateObj.getValidateEndDate() == null) {
            productPriceMapper.update(null, Wrappers.<ProductPriceDO>lambdaUpdate()
                    .set(ProductPriceDO::getValidateEndDate, updateObj.getValidateEndDate())
                    .set(ProductPriceDO::getValidateStartDate, updateObj.getValidateStartDate())
                    .eq(ProductPriceDO::getId, priceId)
            );
        }

        applicationContext.publishEvent(new ProductLineUpdateAutoProcessOverWeightExceptionEvent(priceDO.getProductId(),
                priceDO.getWarehouseLineId(), priceDO.getShippingChannelId(), lineWeightLimitOld, lineWeightLimitNew));

        WarehouseLineDO warehouseLineDO = warehouseLineMapper.selectOne("id", priceDO.getWarehouseLineId());
        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        priceTransportTypeMap.put(priceId, warehouseLineDO.getTransportType());

        List<ProductPriceStepDO> allStepList = buildAllStepList(updateReqVO);
        clearAndSavePriceStepList(Collections.singletonList(updateReqVO.getId()), allStepList, priceTransportTypeMap, false);

        handleOrderException(Collections.singletonList(priceId), true);
    }

    private List<ProductPriceStepDO> buildAllStepList(ProductPriceReqBaseVO productPriceStepBaseVO) {
        List<ProductPriceStepDO> allStepList = new ArrayList<>();
        if(CollectionUtil.isNotEmpty(productPriceStepBaseVO.getFreightPriceStepList())) {
            allStepList.addAll(productPriceStepBaseVO.getFreightPriceStepList());
        }
        if(CollectionUtil.isNotEmpty(productPriceStepBaseVO.getClearancePriceStepList())) {
            allStepList.addAll(productPriceStepBaseVO.getClearancePriceStepList());
        }
        if(CollectionUtil.isNotEmpty(productPriceStepBaseVO.getFullPriceStepList())) {
            allStepList.addAll(productPriceStepBaseVO.getFullPriceStepList());
        }
        return allStepList;
    }

    private List<ProductPricePackagingDO> buildAllPackageList(ProductPriceBaseVO productPriceBaseVO) {
        List<ProductPricePackagingDO> allPackagingList = new ArrayList<>();
        if(CollectionUtil.isNotEmpty(productPriceBaseVO.getFreightPricePackagingList())) {
            productPriceBaseVO.getFreightPricePackagingList().forEach(t -> t.setPriceType(PriceTypeEnum.FREIGHT.getPriceType()));
            allPackagingList.addAll(productPriceBaseVO.getFreightPricePackagingList());
        }
        if(CollectionUtil.isNotEmpty(productPriceBaseVO.getClearancePricePackagingList())) {
            productPriceBaseVO.getClearancePricePackagingList().forEach(t -> t.setPriceType(PriceTypeEnum.CLEARANCE.getPriceType()));
            allPackagingList.addAll(productPriceBaseVO.getClearancePricePackagingList());
        }
        if(CollectionUtil.isNotEmpty(productPriceBaseVO.getFullPricePackagingList())) {
            productPriceBaseVO.getFullPricePackagingList().forEach(t -> t.setPriceType(PriceTypeEnum.FULL_PRICE.getPriceType()));
            allPackagingList.addAll(productPriceBaseVO.getFullPricePackagingList());
        }
        return allPackagingList;
    }

    @Override
    public void batchAddPriceAir(ProductPriceBatchAddPriceAirVo batchAddPriceReqVo, ProductPriceCheckAO productPriceCheckAO) {
        String json = JSON.toJSONString(batchAddPriceReqVo);
        List<ProductDO> productList = productPriceCheckAO.getProductList();
        List<LineChannelVo> lineChannelList = batchAddPriceReqVo.getLineChannelList();
        logAndDelKey(json, productList, lineChannelList);

        List<WarehouseLineDO> warehouseLineList = productPriceCheckAO.getWarehouseLineList();
        String content = productPriceCheckAO.getContent();

        int countBatch = (int) Math.ceil((double) productList.size() / DATA_THRESHOLD);
        List<List<ProductDO>> splitList = Stream.iterate(0, n -> n + 1)
                .limit(countBatch)
                .parallel()
                .map(a -> productList.stream().skip(a * DATA_THRESHOLD).limit(DATA_THRESHOLD).parallel().collect(toList())
                ).collect(toList());
        for (List<ProductDO> itemList : splitList) {
            batchAddPriceAir(batchAddPriceReqVo, warehouseLineList, itemList);
        }
        //发送站内信
        sendMessage(content);
    }

    @Transactional(rollbackFor = Exception.class)
    public void batchAddPriceAir(ProductPriceBatchAddPriceAirVo batchAddPriceReqVo, List<WarehouseLineDO> warehouseLineList, List<ProductDO> productList) {
        long t1 = System.currentTimeMillis();

        List<LineChannelVo> lineChannelList = batchAddPriceReqVo.getLineChannelList();
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());
        Set<Long> channelIdList = lineChannelList.stream()
                .filter(t -> t.getShippingChannelId() != null)
                .map(LineChannelVo::getShippingChannelId)
                .collect(Collectors.toSet());

        Map<Long, String> lineTransportTypeMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, WarehouseLineDO::getTransportType));

        List<Long> idList = productList.stream()
                .map(ProductDO::getId)
                .collect(toList());
        //这个产品这么多线路已有的价格
        List<ProductPriceDO> dbPriceList = productPriceMapper.selectList(new LambdaQueryWrapperX<ProductPriceDO>()
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
                .in(CollectionUtil.isNotEmpty(channelIdList), ProductPriceDO::getShippingChannelId, channelIdList)
                .in(ProductPriceDO::getProductId, idList));


        List<ProductPriceDO> snapshotList = new ArrayList<>();
        List<ProductPriceDO> updateList = new ArrayList<>();

        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        dbPriceList.forEach(t -> priceTransportTypeMap.put(t.getId(), lineTransportTypeMap.get(t.getWarehouseLineId())));

        ProductPriceStepObjectAO freightAO = stepListAddPriceAir(dbPriceList, batchAddPriceReqVo.getFreightPriceStepList(), PriceTypeEnum.FREIGHT, priceTransportTypeMap, snapshotList, updateList);
        ProductPriceStepObjectAO clearanceAO = stepListAddPriceAir(dbPriceList, batchAddPriceReqVo.getClearancePriceStepList(), PriceTypeEnum.CLEARANCE, priceTransportTypeMap, snapshotList, updateList);
        ProductPriceStepObjectAO fullAO = stepListAddPriceAir(dbPriceList, batchAddPriceReqVo.getFullPriceStepList(), PriceTypeEnum.FULL_PRICE, priceTransportTypeMap, snapshotList, updateList);

        //添加修改日志,需要查询特需和阶梯，所以需要在修改之前执行
        batchAddSnapshot(snapshotList);

        if(CollectionUtil.isNotEmpty(updateList)) {
            productPriceMapper.updateBatch(updateList);
        }

        List<ProductPriceStepDO> stepUpdateList = new ArrayList<>();
        List<ProductPriceStepSpecialDO> stepSpecialUpdateList = new ArrayList<>();
        List<ProductPriceStepPackagingDO> stepPackagingUpdateList = new ArrayList<>();

        addStepList(freightAO, stepUpdateList, stepSpecialUpdateList, stepPackagingUpdateList);
        addStepList(clearanceAO, stepUpdateList, stepSpecialUpdateList, stepPackagingUpdateList);
        addStepList(fullAO, stepUpdateList, stepSpecialUpdateList, stepPackagingUpdateList);

        if (CollectionUtil.isNotEmpty(stepUpdateList)) {
            productPriceStepMapper.updateBatch(stepUpdateList);
        }

        if(CollectionUtil.isNotEmpty(stepSpecialUpdateList)) {
            productPriceStepSpecialMapper.updateBatch(stepSpecialUpdateList);
        }

        if(CollectionUtil.isNotEmpty(stepPackagingUpdateList)) {
            productPriceStepPackagingMapper.updateBatch(stepPackagingUpdateList);
        }

        long t2 = System.currentTimeMillis();
        log.info("批量加价路线处理耗时:" + (t2 - t1));
        deleteKey(idList, lineChannelList);
    }

    private void addStepList(ProductPriceStepObjectAO stepObjectAO, List<ProductPriceStepDO> stepUpdateList, List<ProductPriceStepSpecialDO> stepSpecialUpdateList, List<ProductPriceStepPackagingDO> stepPackagingUpdateList) {
        if(stepObjectAO == null) return;
        if(CollectionUtil.isNotEmpty(stepObjectAO.getStepList())) {
            stepUpdateList.addAll(stepObjectAO.getStepList());
        }
        if(CollectionUtil.isNotEmpty(stepObjectAO.getStepSpecialList())) {
            stepSpecialUpdateList.addAll(stepObjectAO.getStepSpecialList());
        }
        if(CollectionUtil.isNotEmpty(stepObjectAO.getStepPackagingList())) {
            stepPackagingUpdateList.addAll(stepObjectAO.getStepPackagingList());
        }
    }

    private ProductPriceStepObjectAO stepListAddPriceAir(List<ProductPriceDO> productPriceList, List<ProductPriceStepDO> stepList, PriceTypeEnum priceTypeEnum, Map<Long, String> priceTransportTypeMap, List<ProductPriceDO> snapshotList, List<ProductPriceDO> updateList) {
        if (stepList == null) return null;

//        stepList = stepList.stream()
//                .filter(t -> t.getTransportPrice() != null ||
//                        t.getClearancePrice() != null ||
//                        t.getAllPrice() != null)
//                .collect(Collectors.toList());

        stepList = stepList.stream()
                .filter(t -> t.getRankNum() != null)
                .collect(toList());

        if (CollectionUtil.isEmpty(stepList)) return null;

        //找出所有价格的所有阶梯数据
        List<ProductPriceStepDO> dbAllStepList = new ArrayList<>();
        List<List<ProductPriceDO>> idPartList = Lists.partition(productPriceList, 1000);
        for (List<ProductPriceDO> priceItemList : idPartList) {
            List<Long> idList = priceItemList.stream()
                    .map(ProductPriceDO::getId)
                    .collect(toList());
            List<ProductPriceStepDO> itemStepList = productPriceStepService.selectList(new LambdaQueryWrapper<ProductPriceStepDO>()
                    .in(ProductPriceStepDO::getProductPriceId, idList)
            );
            dbAllStepList.addAll(itemStepList);
        }

        List<ProductPriceStepPackagingDO> stepPackagingList = stepList.stream()
                .map(ProductPriceStepDO::getPackagingList)
                .filter(CollectionUtil::isNotEmpty)
                .flatMap(List::stream)
                .collect(Collectors.toList());

        List<Long> stepIdList = dbAllStepList.stream()
                .map(ProductPriceStepDO::getId)
                .collect(toList());

        //前端传的需要加价的阶梯数据
        List<ProductPriceStepSpecialDO> allUpdateStepSpecialList = stepList.stream()
                .flatMap(step -> step.getSpecialList().stream())
                .filter(stepSpecialDO -> stepSpecialDO.getTransportPrice() != null
                        || stepSpecialDO.getClearancePrice() != null
                        || stepSpecialDO.getAllPrice() != null)
                .collect(Collectors.toList());
        Map<Long, List<ProductPriceStepSpecialDO>> stepSpecialGroupMap = new HashMap<>();
        if(CollectionUtil.isNotEmpty(allUpdateStepSpecialList)) {
            //数据库查找所有阶梯特需数据
            List<List<Long>> stepIdPartList = Lists.partition(stepIdList, 1000);
            List<ProductPriceStepSpecialDO> allStepSpecialList = new ArrayList<>();
            for (List<Long> stepIdLis : stepIdPartList) {
                List<ProductPriceStepSpecialDO> itemStepSpecialList = productPriceStepSpecialMapper.selectList(new LambdaQueryWrapper<ProductPriceStepSpecialDO>()
                        .in(ProductPriceStepSpecialDO::getProductPriceStepId, stepIdLis)
                );
                allStepSpecialList.addAll(itemStepSpecialList);
            }
            stepSpecialGroupMap = allStepSpecialList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepSpecialDO::getProductPriceStepId));
        }

        Map<Long, List<ProductPriceStepPackagingDO>> stepPackagingGroupMap = new HashMap<>();
        if(CollectionUtil.isNotEmpty(stepPackagingList)) {
            //数据库查找所有阶梯包装数据
            List<List<Long>> stepIdPartList = Lists.partition(stepIdList, 1000);
            List<ProductPriceStepPackagingDO> allStepPackagingList = new ArrayList<>();
            for (List<Long> stepIdLis : stepIdPartList) {
                List<ProductPriceStepPackagingDO> itemStepPackagingList = productPriceStepPackagingMapper.selectList(new LambdaQueryWrapper<ProductPriceStepPackagingDO>()
                        .in(ProductPriceStepPackagingDO::getProductPriceStepId, stepIdLis)
                );
                allStepPackagingList.addAll(itemStepPackagingList);
            }
            stepPackagingGroupMap = allStepPackagingList.stream()
                    .collect(Collectors.groupingBy(ProductPriceStepPackagingDO::getProductPriceStepId));
        }

        Map<Long, List<ProductPriceStepDO>> dbPriceStepMap = dbAllStepList.stream()
                .collect(Collectors.groupingBy(ProductPriceStepDO::getProductPriceId));

        List<ProductPriceStepDO> stepUpdateList = new ArrayList<>();
        List<ProductPriceStepSpecialDO> stepSpecialUpdateList = new ArrayList<>();
        List<ProductPriceStepPackagingDO> stepPackagingUpdateList = new ArrayList<>();

        Map<Integer, ProductPriceStepDO> stepMap = stepList.stream()
                .collect(Collectors.toMap(ProductPriceStepDO::getRankNum, v -> v));

        boolean priceUpdate = false;
        for (ProductPriceDO productPriceDO : productPriceList) {
            Long productPriceId = productPriceDO.getId();
            List<ProductPriceStepDO> productStepList = dbPriceStepMap.get(productPriceId);
            //价格没有阶梯则跳过
            if (CollectionUtil.isEmpty(productStepList)) continue;

            for (ProductPriceStepDO dbStep : productStepList) {
                ProductPriceStepDO updateStep = stepMap.get(dbStep.getRankNum());
                if (updateStep == null) continue;

                //阶梯起始和阶梯顺序一致才更新
                if(updateStep.getStartNum() != null && updateStep.getEndNum() != null) {
                    if(updateStep.getStartNum().intValue() != dbStep.getStartNum().intValue() ||
                            updateStep.getEndNum().intValue() != dbStep.getEndNum().intValue()) {
                        //传了阶梯起始，但是和数据库的不一致则不修改
                        continue;
                    }
                }

                boolean update = addPrice(productPriceDO.getPriceType(), dbStep, updateStep);
                if (update) {
                    priceUpdate = true;
                    stepUpdateList.add(dbStep);
                }

                //前端传的需要加价的阶梯数据
                List<ProductPriceStepSpecialDO> stepSpecialList = updateStep.getSpecialList();
                //校验阶梯特需更新
                if(CollectionUtil.isNotEmpty(stepSpecialList)) {
                    //过滤调空数据
                    stepSpecialList = stepSpecialList.stream()
                            .filter(stepSpecialDO -> stepSpecialDO.getTransportPrice() != null
                                    || stepSpecialDO.getClearancePrice() != null
                                    || stepSpecialDO.getAllPrice() != null)
                            .collect(toList());
                    if(CollectionUtil.isNotEmpty(stepSpecialList)) {
                        Map<String, ProductPriceStepSpecialDO> stepSpecialMap = stepSpecialList.stream()
                                .collect(Collectors.toMap(ProductPriceStepSpecialDO::getSpecialDictType, v -> v));
                        List<ProductPriceStepSpecialDO> dbStepSpecialList = stepSpecialGroupMap.get(dbStep.getId());
                        if(CollectionUtil.isEmpty(dbStepSpecialList)) continue;
                        for (ProductPriceStepSpecialDO dbStepSpecial : dbStepSpecialList) {
                            String specialType = dbStepSpecial.getSpecialDictType();
                            ProductPriceStepSpecialDO updateStepSpecial = stepSpecialMap.get(specialType);
                            if (updateStepSpecial == null) continue;

                            String transportType = priceTransportTypeMap.get(productPriceId);
                            initBasePrice(updateStepSpecial, transportType);
                            boolean stepSpecialUpdate = addPrice(priceTypeEnum.getPriceType(), dbStepSpecial, updateStepSpecial);
                            if (stepSpecialUpdate) {
                                priceUpdate = true;
                                stepSpecialUpdateList.add(dbStepSpecial);
                            }
                        }
                    }
                }

                //阶梯包装费拆成单个包装一条数据后，如果只针对一个包装加价，那就变成了几个包装2条价格，回显问题等等
                //校验阶梯包装价更新
                if(CollectionUtil.isEmpty(stepPackagingList)) continue;
                List<ProductPriceStepPackagingDO> dbStepPackagingList = stepPackagingGroupMap.get(dbStep.getId());
                if(CollectionUtil.isEmpty(dbStepPackagingList)) continue;
                for (ProductPriceStepPackagingDO dbStepPackaging : dbStepPackagingList) {
                    for (ProductPriceStepPackagingDO addPricePackagingDO : stepPackagingList) {
                        boolean stepPackagingUpdate = checkPackageAddPrice(addPricePackagingDO, dbStepPackaging);
                        if(stepPackagingUpdate) {
                            priceUpdate = true;
                            BigDecimal price = dbStepPackaging.getPackagingPrice();
                            price = price.add(addPricePackagingDO.getPackagingPrice());
                            dbStepPackaging.setPackagingPrice(price);
                            stepPackagingUpdateList.add(dbStepPackaging);
                        }
                    }
                }
            }

            if(priceUpdate && snapshotList != null) {
                snapshotList.add(productPriceDO);
            }
            if(priceUpdate && updateList != null) {
                Long loginUserId = SecurityFrameworkUtils.getLoginUserId();
                productPriceDO.setUpdater(loginUserId.toString());
                productPriceDO.setUpdateTime(new Date());
                updateList.add(productPriceDO);
            }
        }

        ProductPriceStepObjectAO objectAO = new ProductPriceStepObjectAO();
        objectAO.setStepList(stepUpdateList);
        objectAO.setStepSpecialList(stepSpecialUpdateList);
        objectAO.setStepPackagingList(stepPackagingUpdateList);
        return objectAO;
    }

    @Override
    public void handleOrderException(Long priceId) {
        List<Long> productPriceIdList = new ArrayList<>();
        productPriceIdList.add(priceId);
        handleOrderException(productPriceIdList, true);
    }

    /**
     *
     * 从商品列表中某一商品进入复制价格
     * @param dto
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public void copyPriceByProductIds(CopyProductPriceDto dto) {
        List<Long> ids = Arrays.stream(dto.getProductPriceIds().split(StrUtil.COMMA)).map(Long::parseLong).collect(Collectors.toList());

        //列出要复制的商品价格
        List<ProductPriceDO> productPriceDOList = this.productPriceMapper.selectList(new LambdaQueryWrapper<ProductPriceDO>().in(ProductPriceDO::getId, ids));
        if(productPriceDOList.size()>0){
            productPriceDOList.forEach(price->{

                ProductPriceAirCreateReqVO createReqVO = new ProductPriceAirCreateReqVO() ;
                BeanUtils.copyProperties(price,createReqVO);
                createReqVO.setProductId(Long.parseLong(dto.getProductId()) );
                LineChannelVo vo = new LineChannelVo();
                vo.setShippingChannelId(price.getShippingChannelId());
                vo.setLineId(price.getWarehouseLineId());
                List<LineChannelVo> lineChannelList = new ArrayList<>();
                lineChannelList.add(vo);
                createReqVO.setLineChannelList(lineChannelList);
                ProductPriceRespVO respVO = this.getProductPriceDetail(price.getId());
                //阶梯运费价格列表
                List<ProductPriceStepDO> freightPriceStepList  = new ArrayList<>();
                //阶梯清关费价格列表
                List<ProductPriceStepDO> clearancePriceStepList = new ArrayList<>() ;
                //阶梯全包价格列表
                List<ProductPriceStepDO> fullPriceStepList  = new ArrayList<>( );
                ProductPriceStepDO freightPriceStep = new ProductPriceStepDO();
                ProductPriceStepDO clear = new ProductPriceStepDO();

                if(CollectionUtil.isEmpty(respVO.getFreightPriceStepList()) ){
                    BeanUtils.copyProperties(price,freightPriceStep);
                    freightPriceStep.setStartNum(BigDecimal.valueOf(0));
                    freightPriceStep.setEndNum(BigDecimal.valueOf(0)) ;
                    freightPriceStep.setRankNum(1);
                    freightPriceStep.setWeightUnit(price.getTransportPriceUnit());
                    freightPriceStep.setProductPriceId(price.getId()) ;
                    freightPriceStepList.add(freightPriceStep);

                    createReqVO.setFreightPriceStepList(freightPriceStepList);
                }else{
                    createReqVO.setFreightPriceStepList(respVO.getFreightPriceStepList()) ;
                }
                if(CollectionUtil.isEmpty(respVO.getClearancePriceStepList()) ){
                    BeanUtils.copyProperties(price,clear);
                    clear.setStartNum(BigDecimal.valueOf(0));
                    clear.setEndNum(BigDecimal.valueOf(0)) ;
                    clear.setRankNum(2);
                    clear.setWeightUnit(price.getClearancePriceUnit());
                    clear.setProductPriceId(price.getId()) ;
                    clearancePriceStepList.add(clear);
                    createReqVO.setClearancePriceStepList(clearancePriceStepList);
                }else{
                    createReqVO.setClearancePriceStepList(respVO.getClearancePriceStepList()) ;
                }



                //ProductDO productDO = checkUpdateStatus(lineChannelList, createReqVO.getProductId());
                ProductDO productDO = productService.getProduct(createReqVO.getProductId());
                        WarehouseLineSearchVO searchVO = new WarehouseLineSearchVO();
                List<WarehouseLineDO> warehouseLineList = warehouseService.openedRouterList(searchVO);

                List<ProductDO> productList = Collections.singletonList(productDO);
                //商品线路保存在redis。更新成功后移除
                batchSaveKeyInCache(productList, lineChannelList);

                createReqVO.setSync(true);
                createReqVO.setBatch(false);
                copyPriceUpdateAir(createReqVO, warehouseLineList, productList,price);
            });
        }


    }

    /**
     * 复制价格
     * @param createReqVO
     * @param warehouseLineList
     * @param productList
     * @param productPrice
     */

    @Transactional(rollbackFor = Exception.class)
    public void copyPriceUpdateAir(ProductPriceAirCreateReqVO createReqVO, List<WarehouseLineDO> warehouseLineList, List<ProductDO> productList,ProductPriceDO productPrice) {
        long t1 = System.currentTimeMillis();
        Long loginUserId = SecurityFrameworkUtils.getLoginUserId();

        checkAndUpdateStepToMainPrice(productPrice, createReqVO);

        List<LineChannelVo> lineChannelList = createReqVO.getLineChannelList();
        Set<Long> lineIdList = lineChannelList.stream()
                .map(LineChannelVo::getLineId)
                .collect(Collectors.toSet());

        Map<Long, String> lineTransportTypeMap = warehouseLineList.stream()
                .collect(Collectors.toMap(WarehouseLineDO::getId, WarehouseLineDO::getTransportType));

        List<Long> idList = productList.stream()
                .map(ProductDO::getId)
                .collect(toList());
        //这个产品这么多线路已有的价格
        List<ProductPriceDO> dbPriceList = productPriceMapper.selectList(new LambdaQueryWrapperX<ProductPriceDO>()
                .in(ProductPriceDO::getWarehouseLineId, lineIdList)
                .in(ProductPriceDO::getProductId, idList));
        Map<String, ProductPriceDO> productLineChannelMap = dbPriceList.stream()
                .collect(Collectors.toMap(t -> t.getProductId() + "-" + t.getWarehouseLineId() + "-" + t.getShippingChannelId(), v -> v, (v1, v2) -> v2));


        List<ProductPriceDO> insertList = new ArrayList<>();
        List<ProductPriceDO> updateList = new ArrayList<>();

        List<ProductPriceDO> snapshotList = new ArrayList<>();

        Date current = new Date();
        for (Long productId : idList) {
            for (LineChannelVo lineChannel : lineChannelList) {
                Long lineId = lineChannel.getLineId();
                Long shippingChannelId = lineChannel.getShippingChannelId();
                String key = productId + "-" + lineId + "-" + shippingChannelId;

                ProductPriceDO dbItem;
                if (productLineChannelMap.containsKey(key)) {
                    dbItem = productLineChannelMap.get(key);

                    //添加日志
                    ProductPriceDO snapshot = ObjectUtil.clone(dbItem);
                    snapshotList.add(snapshot);
                    dbItem.setUpdater(String.valueOf(loginUserId));
                    updateColumn(dbItem, productPrice);
                    updateList.add(dbItem);
                } else {
                    dbItem = ObjectUtil.clone(productPrice);
                    dbItem.setProductId(productId);
                    dbItem.setWarehouseLineId(lineId);
                    dbItem.setShippingChannelId(shippingChannelId);

                    if (dbItem.getNeedBook() == null) {
                        dbItem.setNeedBook(0);
                    }
                    if (dbItem.getPriceType() == null) {
                        dbItem.setPriceType(0);
                    }
                    if (dbItem.getNeedPay() == null) {
                        dbItem.setNeedPay(0);
                    }
                    if (dbItem.getAuditStatus() == null) {
                        dbItem.setAuditStatus(0);
                    }
                    if (dbItem.getPayDelivery() == null) {
                        dbItem.setPayDelivery(0);
                    }
                    if (dbItem.getStepPrice() == null) {
                        dbItem.setStepPrice(0);
                    }
                    if (dbItem.getMinWeight() == null) {
                        dbItem.setMinWeight(BigDecimal.ZERO);
                    }
                    if (dbItem.getMinWeightUnit() == null) {
                        dbItem.setMinWeightUnit(7L);
                    }
                    dbItem.setBlacklist(0);
                    dbItem.setStatus(0);
                    dbItem.setCreateTime(current);
                    insertList.add(dbItem);
                }
                dbItem.setUpdateTime(current);
                String transportType = lineTransportTypeMap.get(lineId);
                initBasePrice(dbItem, transportType);
            }
        }
        if (CollectionUtil.isNotEmpty(insertList)) {
            productPriceMapper.insertBatchSomeColumn(insertList);
        }
        if (CollectionUtil.isNotEmpty(updateList)) {
            productPriceMapper.updateBatch(updateList);
        }
        List<ProductPriceDO> productPriceList = new ArrayList<>();
        productPriceList.addAll(insertList);
        productPriceList.addAll(updateList);
        Map<Long, String> priceTransportTypeMap = new HashMap<>();
        productPriceList.forEach(t -> priceTransportTypeMap.put(t.getId(), lineTransportTypeMap.get(t.getWarehouseLineId())));

        //添加修改日志,需要查询特需和阶梯，所以需要在修改之前执行
        batchAddSnapshot(snapshotList);

        List<Long> productPriceIdList = productPriceList.stream().map(ProductPriceDO::getId).collect(Collectors.toList());

        List<ProductPriceStepDO> allStepList = buildCopyAllStepList(createReqVO);
        clearAndSavePriceStepList(productPriceIdList, allStepList, priceTransportTypeMap, createReqVO.isBatch());

        long t2 = System.currentTimeMillis();
        log.info("路线处理耗时:" + (t2 - t1));
        deleteKey(idList, lineChannelList);

        handleOrderException(productPriceIdList, createReqVO.isSync());
    }

    /**
     * 复制另一商品价格的 阶梯价格
     * @param productPriceStepBaseVO
     * @return
     */
    private List<ProductPriceStepDO> buildCopyAllStepList(ProductPriceReqBaseVO productPriceStepBaseVO) {
        List<ProductPriceStepDO> allStepList = new ArrayList<>();
        if(CollectionUtil.isNotEmpty(productPriceStepBaseVO.getFreightPriceStepList())) {
            allStepList.addAll(productPriceStepBaseVO.getFreightPriceStepList());
        }
        if(CollectionUtil.isNotEmpty(productPriceStepBaseVO.getClearancePriceStepList())) {
            allStepList.addAll(productPriceStepBaseVO.getClearancePriceStepList());
        }
        if(CollectionUtil.isNotEmpty(productPriceStepBaseVO.getFullPriceStepList())) {
            allStepList.addAll(productPriceStepBaseVO.getFullPriceStepList());
        }
        return allStepList;
    }
}
